# coding: utf-8
"""Tools for computing derivatives by finite differences."""
import numpy as np

from monty.collections import dict2namedtuple

__all__ = [
    "finite_diff",
]


def rearr(array):
    return np.array(array, dtype=float)

# This table contains the coefficients of the central differences, for several order of accuracy: [1]
# See http://en.wikipedia.org/wiki/Finite_difference_coefficients
# Derivative Accuracy -4 -3 -2 -1 0 1 2 3 4


central_fdiff_weights = {
1: {
    2: rearr([-1/2, 0, 1/2]),
    4: rearr([1/12, -2/3, 0, 2/3, -1/12]),
    6: rearr([-1/60, 3/20, -3/4, 0, 3/4, -3/20, 1/60]),
    8: rearr([1/280, -4/105, 1/5, -4/5, 0, 4/5, -1/5, 4/105, -1/280]),
    },
2: {
    2: rearr([1, -2, 1]),
    4: rearr([-1/12, 4/3, -5/2, 4/3, -1/12]),
    6: rearr([1/90, -3/20, 3/2, -49/18, 3/2, -3/20, 1/90]),
    8: rearr([-1/560, 8/315, -1/5, 8/5, -205/72, 8/5, -1/5, 8/315, -1/560]),
    },
3: {
    2: rearr([-1/2, 1, 0, -1, 1/2]),
    4: rearr([1/8, -1, 13/8, 0, -13/8, 1, -1/8]),
    6: rearr([-7/240, 3/10, -169/120, 61/30, 0, -61/30, 169/120, -3/10, 7/240]),
    },
4: {
    2: rearr([1, -4, 6, -4, 1]),
    4: rearr([-1/6, 2, -13/2, 28/3, -13/2, 2, -1/6]),
    6: rearr([7/240, -2/5, 169/60, -122/15, 91/8, -122/15, 169/60, -2/5, 7/240]),
    },
}


#Derivative  Accuracy   0 1 2 3 4 5 6 7 8
forward_fdiff_weights = {
1: {
    1: rearr([-1, 1]),
    2: rearr([-3/2, 2, -1/2]),
    3: rearr([-11/6, 3, -3/2, 1/3]),
    4: rearr([-25/12, 4, -3, 4/3, -1/4]),
    5: rearr([-137/60, 5, -5, 10/3, -5/4, 1/5]),
    6: rearr([-49/20, 6, -15/2, 20/3, -15/4, 6/5, -1/6]),
},
2: {
    1: rearr([1, -2, 1]),
    2: rearr([2, -5, 4, -1]),
    3: rearr([35/12, -26/3, 19/2, -14/3, 11/12]),
    4: rearr([15/4,  -77/6, 107/6, -13, 61/12, -5/6]),
    5: rearr([203/45, -87/5, 117/4, -254/9, 33/2, -27/5, 137/180]),
    6: rearr([469/90, -223/10, 879/20, -949/18, 41, -201/10, 1019/180, -7/10]),
},
3: {
    1: rearr([-1, 3, -3, 1]),
    2: rearr([-5/2, 9, -12, 7, -3/2]),
    3: rearr([-17/4, 71/4, -59/2, 49/2, -41/4, 7/4]),
    4: rearr([-49/8, 29, -461/8, 62, -307/8, 13, -15/8]),
    5: rearr([-967/120, 638/15, -3929/40, 389/3, -2545/24, 268/5, -1849/120, 29/15]),
    6: rearr([-801/80, 349/6, -18353/120, 2391/10, -1457/6, 4891/30, -561/8, 527/30, -469/240]),
},
4: {
    1: rearr([1, -4, 6, -4, 1]),
    2: rearr([3, -14, 26, -24, 11, -2]),
    3: rearr([35/6, -31, 137/2, -242/3, 107/2, -19, 17/6]),
    4: rearr([28/3, -111/2, 142, -1219/6, 176, -185/2,  82/3, -7/2]),
    5: rearr([1069/80, -1316/15, 15289/60, -2144/5, 10993/24, -4772/15, 2803/20, -536/15, 967/240]),
},
}
del rearr

# To get the coefficients of the backward approximations,
# give all odd derivatives listed in the table the opposite sign,
# whereas for even derivatives the signs stay the same.
backward_fdiff_weights = d = {}

for ord, v in forward_fdiff_weights.items():
    d[ord] = {}
    for accuracy, weights in v.items():
        d[ord][accuracy] = ((-1)**ord) * weights[-1::-1]


def finite_diff(arr, h, order=1, acc=4, index=None):
    """
    Compute the derivative of order `order` by finite difference.
    For each point in arr, the function tries to use central differences
    and fallbacks to forward/backward approximations for points that are close to the extrema.
    Note that high accuracy levels can fail and raise `ValueError` if not enough points are available in `arr`.

    Args:
        arr: Input array with y-values.
        h: Spacing along x
        order: Derivative order
        acc: accuracy level.
        index: If not None, gives the index of the single element in arr where the derivative is wanted.
            In this case a namedtuple with the derivative, the number of points used and the mode is returned

    Return:
        numpy array or (value, npts, mode) if index is not None .
    """
    arr = np.asarray(arr)

    if np.iscomplexobj(arr):
        raise ValueError("Derivatives of complex functions are not supported!")

    # Retrieve weights.
    try:
        centr_ws = central_fdiff_weights[order][acc]
    except KeyError:
        raise ValueError("Centeral diff weights for order: %s, and accuracy: %s are missing!" % (order, acc))

    npsum = np.sum
    ders = np.empty(arr.shape)
    n = len(arr)
    cpad = len(centr_ws) // 2

    for i in range(n):
        if index is not None and i != index: continue
        start = i - cpad
        stop = i + cpad + 1

        if start >= 0 and stop <= n:
            # Can do central difference.
            ders[i] = npsum(centr_ws * arr[start:stop])
            npts = len(centr_ws)
            mode = "central"

        elif start < 0:
            # Try forward.
            forw_ws = forward_fdiff_weights[order][acc]
            stop = i + len(forw_ws)
            if stop > n:
                raise ValueError(
                        ("\n\tDon't have enough points for index: %s in array of lenght: %s\n" +
                         "\tto compute forward finite difference with order: %s, and acc: %s (num_weights: %s)\n" +
                         "\tDecrease acc or increase the number of sampling points.") % (i, n, order, acc, len(forw_ws)))
            ders[i] = npsum(forw_ws * arr[i:stop])
            npts = len(forw_ws)
            mode = "forward"

        elif stop > n:
            # Try backward.
            back_ws = backward_fdiff_weights[order][acc]
            start = i - len(back_ws) + 1
            if start < 0:
                raise ValueError(
                    ("\n\tDon't have enough points for index: %s in array of length: %s\n" +
                    "\tto compute backward finite difference with order: %s, and acc: %s (num_weights: %s)\n" +
                    "\tDecrease acc or increase the number of sampling points.") % (i, n, order, acc, len(back_ws)))
            ders[i] = npsum(back_ws * arr[start:i+1])
            npts = len(back_ws)
            mode = "backward"

    if index is None:
        return ders / (h ** order)
    else:
        return dict2namedtuple(value=ders[index] / (h ** order), npts=npts, mode=mode)



#!/usr/bin/env python
r"""
Gruneisen parameters with DFPT phonons and finite difference
============================================================

This script compute the Grüneisen parameters (derivative of frequencies wrt Volume)
using finite differences and the phonons obtained with the DFPT part of Abinit.
The Anaddb input file needed to compute Grüneisen parameters will be generated
in the outdata directory of the flow.

It is necessary to run three DFPT phonon calculations.
One is calculated at the equilibrium volume and the remaining two are calculated
at the slightly larger volume and smaller volume than the equilibrium volume.
The unitcells at these volumes have to be fully relaxed under the constraint of each volume.
"""
import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata
from abipy import flowtk


def build_flow(options):
    """
    Create a `Flow` for Grüneisen calculations:
    Three relaxations at fixed volume followed by phonon calculation on a q-mesh.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=options.workdir)

    # This function constructs the input files for the phonon calculation:
    # GS input + the input files for the phonon calculation.
    pseudos = abidata.pseudos("14si.pspnc")
    structure = abidata.structure_from_ucell("Si")

    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    # Global variables used both for the GS and the DFPT run.
    gs_inp.set_vars(
        nband=4,
        ecut=2.0,
        ngkpt=[4, 4, 4],
        nshiftk=4,
        shiftk=[0.0, 0.0, 0.5,   # This gives the usual fcc Monkhorst-Pack grid
                0.0, 0.5, 0.0,
                0.5, 0.0, 0.0,
                0.5, 0.5, 0.5],
        diemac=12.0,
        #iomode=3,
        tolvrs=1.0e-18,
    )

    # NB: k-mesh in gs_inp and ngqpt q-mesh must be commensurate.
    from abipy.flowtk.gruneisen import GruneisenWork
    voldelta = gs_inp.structure.volume * 0.02
    work = GruneisenWork.from_gs_input(gs_inp, voldelta, ngqpt=[2, 2, 2], with_becs=False)
    flow.register_work(work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).plot_networkx(tight_layout=True)


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
G0W0 spectral function
======================

This examples shows how to plot the G0W0 spectral functions A(w) at the gamma point.
See also lesson tgw2_4
"""
import abipy.data as abidata
from abipy.abilab import abiopen

# Open the file with the GW results
sigres = abiopen(abidata.ref_file("al_g0w0_sigmaw_SIGRES.nc"))

# Plot A(w) for the first spin, the gamma point, and all bands
sigres.plot_spectral_functions()

# Only bands in [0, 1, 2]
sigres.plot_spectral_functions(spin=0, kpoint=(0, 0, 0), include_bands=range(0, 3))

sigres.close()



# coding: utf-8
"""
Object to analyze the results stored in the V1SYM.nc file (mainly for debugging purposes)
"""
import numpy as np

from collections import OrderedDict
from monty.string import marquee
from monty.functools import lazy_property
from abipy.tools.plotting import add_fig_kwargs, get_axarray_fig_plt
from abipy.core.mixins import AbinitNcFile, Has_Structure, NotebookWriter
from abipy.core.kpoints import KpointList, Kpoint
from abipy.iotools import ETSF_Reader
from abipy.tools import duck


class V1symFile(AbinitNcFile, Has_Structure, NotebookWriter):

    def __init__(self, filepath):
        super().__init__(filepath)
        self.reader = r = ETSF_Reader(filepath)
        # Read dimensions.
        self.nfft = r.read_dimvalue("nfft")
        self.nspden = r.read_dimvalue("nspden")
        self.natom3 = len(self.structure) * 3
        self.symv1scf = r.read_value("symv1scf")
        # Read FFT mesh.
        #self.ngfft = r.read_value("ngfft")

    @lazy_property
    def structure(self):
        """|Structure| object."""
        return self.reader.read_structure()

    @lazy_property
    def pertsy_qpt(self):
        """
        Determine the symmetrical perturbations. Meaning of pertsy:

        0 for non-target perturbations.
        1 for basis perturbations.
        -1 for perturbations that can be found from basis perturbations.
        """
        # Fortran array: nctkarr_t("pertsy_qpt", "int", "three, mpert, nqpt")))
        return self.reader.read_value("pertsy_qpt")

    def close(self):
        self.reader.close()

    @lazy_property
    def params(self):
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        return {}

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        lines = []; app = lines.append
        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app("symv1scf: %s" % self.symv1scf)

        return "\n".join(lines)

    @lazy_property
    def qpoints(self):
        return KpointList(self.structure.reciprocal_lattice, frac_coords=self.reader.read_value("qpts"))

    def _find_iqpt_qpoint(self, qpoint):
        if duck.is_intlike(qpoint):
            iq = qpoint
            qpoint = self.qpoints[iq]
        else:
            qpoint = Kpoint.as_kpoint(qpoint, self.structure.reciprocal_lattice)
            iq = self.qpoints.index(qpoint)

        return iq, qpoint

    def read_v1_at_iq(self, key, iq, reshape_nfft_nspden=False):
        # Fortran array ("two, nfft, nspden, natom3, nqpt")
        v1 = self.reader.read_variable(key)[iq]
        v1 = v1[..., 0] + 1j * v1[..., 1]
        # reshape (nspden, nfft) dims because we are not interested in the spin dependence.
        if reshape_nfft_nspden: v1 = np.reshape(v1, (self.natom3, self.nspden * self.nfft))
        return v1

    @add_fig_kwargs
    def plot_diff_at_qpoint(self, qpoint=0, fontsize=8, **kwargs):
        """
        Args:
            qpoint:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        iq, qpoint = self._find_iqpt_qpoint(qpoint)

        # complex arrays with shape: (natom3, nspden * nfft)
        origin_v1 = self.read_v1_at_iq("origin_v1scf", iq, reshape_nfft_nspden=True)
        symm_v1 = self.read_v1_at_iq("recons_v1scf", iq, reshape_nfft_nspden=True)

        num_plots, ncols, nrows = self.natom3, 3, self.natom3 // 3
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)

        for nu, ax in enumerate(ax_list.ravel()):
            idir = nu % 3
            ipert = (nu - idir) // 3

            # l1_rerr(f1, f2) = \int |f1 - f2| dr / (\int |f2| dr
            abs_diff = np.abs(origin_v1[nu] - symm_v1[nu])
            l1_rerr = np.sum(abs_diff) / np.sum(np.abs(origin_v1[nu]))

            stats = OrderedDict([
                ("max", abs_diff.max()),
                ("min", abs_diff.min()),
                ("mean", abs_diff.mean()),
                ("std", abs_diff.std()),
                ("L1_rerr", l1_rerr),
            ])

            xs = np.arange(len(abs_diff))
            ax.hist(abs_diff, facecolor='g', alpha=0.75)
            ax.grid(True)
            ax.set_title("idir: %d, iat: %d, pertsy: %d" % (idir, ipert, self.pertsy_qpt[iq, ipert, idir]),
                         fontsize=fontsize)

            ax.axvline(stats["mean"], color='k', linestyle='dashed', linewidth=1)
            _, max_ = ax.get_ylim()
            ax.text(0.7, 0.7, "\n".join("%s = %.1E" % item for item in stats.items()),
                    fontsize=fontsize, horizontalalignment='center', verticalalignment='center',
                    transform=ax.transAxes)

        fig.suptitle("qpoint: %s" % repr(qpoint))
        return fig

    @add_fig_kwargs
    def plot_pots_at_qpoint(self, qpoint=0, fontsize=8, **kwargs):
        """
        Args:
            qpoint:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        iq, qpoint = self._find_iqpt_qpoint(qpoint)

        # complex arrays with shape: (natom3, nspden * nfft)
        origin_v1 = self.read_v1_at_iq("origin_v1scf", iq, reshape_nfft_nspden=True)
        symm_v1 = self.read_v1_at_iq("recons_v1scf", iq, reshape_nfft_nspden=True)

        num_plots, ncols, nrows = self.natom3, 3, self.natom3 // 3
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)

        natom = len(self.structure)
        xs = np.arange(self.nspden * self.nfft)
        for nu, ax in enumerate(ax_list.ravel()):
            idir = nu % 3
            ipert = (nu - idir) // 3

            # l1_rerr(f1, f2) = \int |f1 - f2| dr / (\int |f2| dr
            abs_diff = np.abs(origin_v1[nu] - symm_v1[nu])
            l1_rerr = np.sum(abs_diff) / np.sum(np.abs(origin_v1[nu]))

            stats = OrderedDict([
                ("max", abs_diff.max()),
                ("min", abs_diff.min()),
                ("mean", abs_diff.mean()),
                ("std", abs_diff.std()),
                ("L1_rerr", l1_rerr),
            ])

            ax.grid(True)
            ax.set_title("idir: %d, iat: %d, pertsy: %d" % (idir, ipert, self.pertsy_qpt[iq, ipert, idir]),
                         fontsize=fontsize)
            # Plot absolute error
            #ax.plot(xs, abs_diff, linestyle="-", color="red", alpha=1.0, label="Abs diff" if nu == 0 else None)

            # Plot absolute values
            #ax.plot(xs, np.abs(origin_v1[nu]), linestyle="--", color="red", alpha=0.4, label="Origin" if nu == 0 else None)
            #ax.plot(xs, -np.abs(symm_v1[nu]), linestyle="--", color="blue", alpha=0.4, label="-Symm" if nu == 0 else None)

            # Plot real and imag
            #ax.plot(xs, origin_v1[nu].real, linestyle="--", color="red", alpha=0.4, label="Re Origin" if nu == 0 else None)
            #ax.plot(xs, -symm_v1[nu].real, linestyle="--", color="blue", alpha=0.4, label="Re Symm" if nu == 0 else None)

            data = np.angle(origin_v1[nu], deg=True) - np.angle(symm_v1[nu], deg=True)
            #data = data[abs_diff > stats["mean"]]
            data = data[np.abs(origin_v1[nu]) > 1e-5]
            ax.plot(np.arange(len(data)), data,
                    linestyle="--", color="red", alpha=0.4, label="diff angle degrees" if nu == 0 else None)

            #ax.plot(xs, origin_v1[nu].real, linestyle="--", color="red", alpha=0.4, label="Re Origin" if nu == 0 else None)
            #ax.plot(xs, -symm_v1[nu].real, linestyle="--", color="blue", alpha=0.4, label="Re Symm" if nu == 0 else None)

            #ax.plot(xs, origin_v1[nu].real - symm_v1[nu].real, linestyle="--", color="red", alpha=0.4,
            #        label="Re Origin" if nu == 0 else None)

            #ax.plot(xs, origin_v1[nu].imag, linestyle=":", color="red", alpha=0.4, label="Imag Origin" if nu == 0 else None)
            #ax.plot(xs, -symm_v1[nu].imag, linestyle=":", color="blue", alpha=0.4, label="Imag Symm" if nu == 0 else None)

            #ax.plot(xs, origin_v1[nu].imag - symm_v1[nu].imag, linestyle="--", color="blue", alpha=0.4,
            #        label="Re Origin" if nu == 0 else None)

            if nu == 0:
                ax.set_ylabel(r"Abs diff")
                ax.legend(loc="best", fontsize=fontsize, shadow=True)
            if ipert == natom - 1:
                ax.set_xlabel(r"FFT index")

            #ax.axvline(stats["mean"], color='k', linestyle='dashed', linewidth=1)
            _, max_ = ax.get_ylim()
            ax.text(0.7, 0.7, "\n".join("%s = %.1E" % item for item in stats.items()),
                    fontsize=fontsize, horizontalalignment='center', verticalalignment='center',
                    transform=ax.transAxes)

            #ax2 = ax.twinx()
            #rerr = 100 * abs_diff / np.abs(origin_v1[nu])
            #ax2.plot(xs, rerr, linestyle="--", color="blue", alpha=0.4,
            #          label=r"|V_{\mathrm{origin}}|" if nu == 0 else None)

        fig.suptitle("qpoint: %s" % repr(qpoint))
        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        maxnq = 3
        for iq, qpoint in enumerate(self.qpoints):
            if iq > maxnq:
                print("Only the first %d q-points are show..." % maxnq)
                break
            #yield self.plot_diff_at_qpoint(qpoint=iq, **kwargs, show=False)
            yield self.plot_pots_at_qpoint(qpoint=iq, **kwargs, show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
        ])

        for iq, qpoint in enumerate(self.qpoints):
            nb.cells.append(nbv.new_code_cell("ncfile.plot_diff_at_qpoint(qpoint=%d);" % iq))
            #nb.cells.append(nbv.new_code_cell("ncfile.plot_diff_at_qpoint(qpoint=%d);" % iq))

        return self._write_nb_nbpath(nb, nbpath)



# coding: utf-8
"""History file with structural relaxation results."""
from __future__ import annotations

import os
import numpy as np
import pandas as pd
import pymatgen.core.units as units
import abipy.core.abinit_units as abu

from collections import OrderedDict
from monty.functools import lazy_property
from monty.collections import AttrDict
from monty.string import marquee, list_strings
from pymatgen.core.periodic_table import Element
from pymatgen.analysis.structure_analyzer import RelaxationAnalyzer
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_visible, get_figs_plotly, \
    get_fig_plotly, add_plotly_fig_kwargs, plotlyfigs_to_browser, push_to_chart_studio, PlotlyRowColDesc, plotly_set_lims, \
    latex_greek_2unicode
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, NotebookWriter
from abipy.abio.robots import Robot
from abipy.iotools import ETSF_Reader
from abipy.tools.typing import Figure
from abipy.core.structure import Structure


class HistFile(AbinitNcFile, NotebookWriter):
    """
    File with the history of a structural relaxation or molecular dynamics calculation.

    Usage example:

    .. code-block:: python

        with HistFile("foo_HIST") as hist:
            hist.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: HistFile
    """
    @classmethod
    def from_file(cls, filepath: str) -> HistFile:
        """Initialize the object from a netcdf_ file"""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = self.r = HistReader(filepath)

    def close(self) -> None:
        """Close the file."""
        self.r.close()

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        return {}

    def __str__(self) -> str:
        return self.to_string()

    # TODO: Add more metadata.
    #@lazy_property
    #def nsppol(self):
    #    """Number of independent spins."""
    #    return self.r.read_dimvalue("nsppol")

    #@lazy_property
    #def nspden(self):
    #    """Number of independent spin densities."""
    #    return self.r.read_dimvalue("nspden")

    #@lazy_property
    #def nspinor(self):
    #    """Number of spinor components."""
    #    return self.r.read_dimvalue("nspinor")

    @lazy_property
    def final_energy(self) -> float:
        """Total energy in eV of the last iteration."""
        return self.etotals[-1]

    @lazy_property
    def final_pressure(self) -> float:
        """Final pressure in Gpa."""
        cart_stress_tensors, pressures = self.r.read_cart_stress_tensors()
        return pressures[-1]

    #@lazy_property
    #def final_max_force(self):

    def get_fstats_dict(self, step) -> AttrDict:
        """
        Return |AttrDict| with stats on the forces at the given ``step``.
        """
        # [time, natom, 3]
        var = self.r.read_variable("fcart")
        forces = units.ArrayWithUnit(var[step], "Ha bohr^-1").to("eV ang^-1")
        fmods = np.array([np.linalg.norm(force) for force in forces])

        return AttrDict(
            fmin=fmods.min(),
            fmax=fmods.max(),
            fmean=fmods.mean(),
            fstd=fmods.std(),
            drift=np.linalg.norm(forces.sum(axis=0)),
        )

    def to_string(self, verbose=0, title=None) -> str:
        """String representation."""
        lines = []; app = lines.append
        if title is not None: app(marquee(title, mark="="))

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.initial_structure.to_string(verbose=verbose, title="Initial Structure"))
        app("")
        app("Number of relaxation steps performed: %d" % self.num_steps)
        app(self.final_structure.to_string(verbose=verbose, title="Final structure"))
        app("")

        an = self.get_relaxation_analyzer()
        app("Volume change in percentage: %.2f%%" % (an.get_percentage_volume_change() * 100))
        d = an.get_percentage_lattice_parameter_changes()
        vals = tuple(d[k] * 100 for k in ("a", "b", "c"))
        app("Percentage lattice parameter changes:\n\ta: %.2f%%, b: %.2f%%, c: %2.f%%" % vals)
        #an.get_percentage_bond_dist_changes(max_radius=3.0)
        app("")

        cart_stress_tensors, pressures = self.r.read_cart_stress_tensors()
        app("Stress tensor (Cartesian coordinates in GPa):\n%s" % cart_stress_tensors[-1])
        app("Pressure: %.3f [GPa]" % pressures[-1])

        return "\n".join(lines)

    @property
    def num_steps(self) -> int:
        """Number of iterations performed."""
        return self.r.num_steps

    @lazy_property
    def steps(self) -> list:
        """Step indices."""
        return list(range(self.num_steps))

    @property
    def initial_structure(self) -> Structure:
        """The initial |Structure|."""
        return self.structures[0]

    @property
    def final_structure(self) -> Structure:
        """The |Structure| of the last iteration."""
        return self.structures[-1]

    @lazy_property
    def structures(self) -> list[Structure]:
        """List of |Structure| objects at the different steps."""
        return self.r.read_all_structures()

    @lazy_property
    def etotals(self) -> np.ndarray:
        """|numpy-array| with total energies in eV at the different steps."""
        return self.r.read_eterms().etotals

    def get_relaxation_analyzer(self) -> RelaxationAnalyzer:
        """
        Return a pymatgen :class:`RelaxationAnalyzer` object to analyze the relaxation in a calculation.
        """
        return RelaxationAnalyzer(self.initial_structure, self.final_structure)

    def to_xdatcar(self, filepath=None, groupby_type=True, to_unit_cell=False, **kwargs) -> Xdatcar:
        """
        Return Xdatcar pymatgen object. See write_xdatcar for the meaning of arguments.

        Args:
            to_unit_cell (bool): Whether to translate sites into the unit cell.
            kwargs: keywords arguments passed to Xdatcar constructor.
        """
        from pymatgen.io.vasp.outputs import Xdatcar

        filepath = self.write_xdatcar(filepath=filepath, groupby_type=groupby_type,
                                      to_unit_cell=to_unit_cell, overwrite=True)

        return Xdatcar(filepath, **kwargs)

    def write_xdatcar(self, filepath="XDATCAR", groupby_type=True, overwrite=False, to_unit_cell=False) -> str:
        """
        Write Xdatcar file with unit cell and atomic positions to file ``filepath``.

        Args:
            filepath: Xdatcar filename. If None, a temporary file is created.
            groupby_type: If True, atoms are grouped by type. Note that this option
                may change the order of the atoms. This option is needed because
                there are post-processing tools (e.g. ovito) that do not work as expected
                if the atoms in the structure are not grouped by type.
            overwrite: raise RuntimeError, if False and filepath exists.
            to_unit_cell (bool): Whether to translate sites into the unit cell.

        Return:
            path to Xdatcar file.
        """
        # This library takes 13s to import on HPC (07/02/24) so moved to class method instead of header
        from pymatgen.io.vasp.outputs import Xdatcar

        if filepath is not None and os.path.exists(filepath) and not overwrite:
            raise RuntimeError("Cannot overwrite pre-existing file `%s`" % filepath)

        if filepath is None:
            import tempfile
            fd, filepath = tempfile.mkstemp(text=True, suffix="_XDATCAR")

        # int typat[natom], double znucl[npsp]
        # NB: typat is double in the HIST.nc file
        typat = self.r.read_value("typat").astype(int)
        znucl = self.r.read_value("znucl")
        ntypat = self.r.read_dimvalue("ntypat")
        num_pseudos = self.r.read_dimvalue("npsp")
        if num_pseudos != ntypat:
            raise NotImplementedError("Alchemical mixing is not supported, num_pseudos != ntypat")
        #print("znucl:", znucl, "\ntypat:", typat)

        symb2pos = OrderedDict()
        symbols_atom = []
        for iatom, itype in enumerate(typat):
            itype = itype - 1
            symbol = Element.from_Z(int(znucl[itype])).symbol
            if symbol not in symb2pos: symb2pos[symbol] = []
            symb2pos[symbol].append(iatom)
            symbols_atom.append(symbol)

        if not groupby_type:
            group_ids = np.arange(self.r.natom)
        else:
            group_ids = []
            for pos_list in symb2pos.values():
                group_ids.extend(pos_list)
            group_ids = np.array(group_ids, dtype=int)

        comment = " %s\n" % self.initial_structure.formula
        with open(filepath, "wt") as fh:
            # comment line  + scaling factor set to 1.0
            fh.write(comment)
            fh.write("1.0\n")
            for vec in self.initial_structure.lattice.matrix:
                fh.write("%.12f %.12f %.12f\n" % (vec[0], vec[1], vec[2]))
            if not groupby_type:
                fh.write(" ".join(symbols_atom) + "\n")
                fh.write("1 " * len(symbols_atom) + "\n")
            else:
                fh.write(" ".join(symb2pos.keys()) + "\n")
                fh.write(" ".join(str(len(p)) for p in symb2pos.values()) + "\n")

            # Write atomic positions in reduced coordinates.
            xred_list = self.r.read_value("xred")
            if to_unit_cell:
                xred_list = xred_list % 1

            for step in range(self.num_steps):
                fh.write("Direct configuration= %d\n" % (step + 1))
                frac_coords = xred_list[step, group_ids]
                for fs in frac_coords:
                    fh.write("%.12f %.12f %.12f\n" % (fs[0], fs[1], fs[2]))

        return filepath

    def visualize(self, appname="ovito", to_unit_cell=False):  # pragma: no cover
        """
        Visualize the crystalline structure with visualizer.
        See :class:`Visualizer` for the list of applications and formats supported.

        Args:
            to_unit_cell (bool): Whether to translate sites into the unit cell.
        """
        if appname == "mayavi": return self.mayaview()

        # Get the Visualizer subclass from the string.
        from abipy.iotools import Visualizer
        visu = Visualizer.from_name(appname)
        if visu.name != "ovito":
            raise NotImplementedError("visualizer: %s" % visu.name)

        filepath = self.write_xdatcar(filepath=None, groupby_type=True, to_unit_cell=to_unit_cell)

        return visu(filepath)()
        #if options.trajectories:
        #    hist.mvplot_trajectories()
        #else:
        #    hist.mvanimate()

    def plot_ax(self, ax, what, fontsize=8, **kwargs) -> None:
        """
        Helper function to plot quantity ``what`` on axis ``ax`` with matplotlib.

        Args:
            fontsize: fontsize for legend.
            kwargs are passed to matplotlib plot method.
        """
        label = None
        if what == "energy":
            # Total energy in eV.
            marker = kwargs.pop("marker", "o")
            label = kwargs.pop("label", "Energy")
            ax.plot(self.steps, self.etotals, label=label, marker=marker, **kwargs)
            ax.set_ylabel('Energy (eV)')

        elif what == "abc":
            # Lattice parameters.
            mark = kwargs.pop("marker", None)
            markers = ["o", "^", "v"] if mark is None else 3 * [mark]
            for i, label in enumerate(["a", "b", "c"]):
                ax.plot(self.steps, [s.lattice.abc[i] for s in self.structures], label=label,
                        marker=markers[i], **kwargs)
            ax.set_ylabel("abc (A)")

        elif what in ("a", "b", "c"):
            i = ("a", "b", "c").index(what)
            marker = kwargs.pop("marker", None)
            if marker is None:
                marker = {"a": "o", "b": "^", "c": "v"}[what]
            label = kwargs.pop("label", what)
            ax.plot(self.steps, [s.lattice.abc[i] for s in self.structures], label=label,
                    marker=marker, **kwargs)
            ax.set_ylabel('%s (A)' % what)

        elif what == "angles":
            # Lattice Angles
            mark = kwargs.pop("marker", None)
            markers = ["o", "^", "v"] if mark is None else 3 * [mark]
            for i, label in enumerate(["alpha", "beta", "gamma"]):
                ax.plot(self.steps, [s.lattice.angles[i] for s in self.structures], label=label,
                        marker=markers[i], **kwargs)
            ax.set_ylabel(r"$\alpha\beta\gamma$ (degree)")

        elif what in ("alpha", "beta", "gamma"):
            i = ("alpha", "beta", "gamma").index(what)
            marker = kwargs.pop("marker", None)
            if marker is None:
                marker = {"alpha": "o", "beta": "^", "gamma": "v"}[what]

            label = kwargs.pop("label", what)
            ax.plot(self.steps, [s.lattice.angles[i] for s in self.structures], label=label,
                    marker=marker, **kwargs)
            ax.set_ylabel(r"$\%s$ (degree)" % what)

        elif what == "volume":
            marker = kwargs.pop("marker", "o")
            ax.plot(self.steps, [s.lattice.volume for s in self.structures], marker=marker, **kwargs)
            ax.set_ylabel(r'$V\, (A^3)$')

        elif what == "pressure":
            stress_cart_tensors, pressures = self.r.read_cart_stress_tensors()
            marker = kwargs.pop("marker", "o")
            label = kwargs.pop("label", "P")
            ax.plot(self.steps, pressures, label=label, marker=marker, **kwargs)
            ax.set_ylabel('P (GPa)')

        elif what == "forces":
            forces_hist = self.r.read_cart_forces()
            fmin_steps, fmax_steps, fmean_steps, fstd_steps = [], [], [], []
            for step in range(self.num_steps):
                forces = forces_hist[step]
                fmods = np.sqrt([np.dot(force, force) for force in forces])
                fmean_steps.append(fmods.mean())
                fstd_steps.append(fmods.std())
                fmin_steps.append(fmods.min())
                fmax_steps.append(fmods.max())

            mark = kwargs.pop("marker", None)
            markers = ["o", "^", "v", "X"] if mark is None else 4 * [mark]
            ax.plot(self.steps, fmin_steps, label="min |F|", marker=markers[0], **kwargs)
            ax.plot(self.steps, fmax_steps, label="max |F|", marker=markers[1], **kwargs)
            ax.plot(self.steps, fmean_steps, label="mean |F|", marker=markers[2], **kwargs)
            ax.plot(self.steps, fstd_steps, label="std |F|", marker=markers[3], **kwargs)
            label = "std |F|"
            ax.set_ylabel('F stats (eV/A)')

        else:
            raise ValueError(f"Invalid value for {what=}")

        ax.set_xlabel('Step')
        ax.grid(True)
        if label is not None:
            ax.legend(loc='best', fontsize=fontsize, shadow=True)

    def plotly_traces(self, fig, what, rcd=None, fontsize=8, showlegend=False, **kwargs):
        """
        Helper function to plot quantity ``what`` on figure ``fig`` with plotly.

        Args:
            rcd: If ``fig`` has subplots, ``rcd`` is used to add traces on these subplots.
            fontsize: fontsize for legend.
            kwargs are passed to fig.add_scatter method.
        """
        rcd = PlotlyRowColDesc.from_object(rcd)
        ply_row, ply_col = rcd.ply_row, rcd.ply_col

        label = None
        if what == "energy":
            # Total energy in eV.
            marker = kwargs.pop("marker", 0)
            label = kwargs.pop("label", "Energy")
            fig.add_scatter(x=self.steps, y=self.etotals, mode='lines+markers', name=label, marker_symbol=marker,
                            row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = 'Energy (eV)'

        elif what == "abc":
            # Lattice parameters.
            mark = kwargs.pop("marker", None)
            markers = [0, 5, 6] if mark is None else 3 * [mark]
            for i, label in enumerate(["a", "b", "c"]):
                fig.add_scatter(x=self.steps, y=[s.lattice.abc[i] for s in self.structures], mode='lines+markers',
                                name=label, marker_symbol=markers[i], row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = "abc (A)"

        elif what in ("a", "b", "c"):
            i = ("a", "b", "c").index(what)
            marker = kwargs.pop("marker", None)
            if marker is None:
                marker = {"a": 0, "b": 5, "c": 6}[what]
            label = kwargs.pop("label", what)
            fig.add_scatter(x=self.steps, y=[s.lattice.abc[i] for s in self.structures], mode='lines+markers',
                            name=label, marker_symbol=marker, row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = '%s (A)' % what

        elif what == "angles":
            # Lattice Angles
            mark = kwargs.pop("marker", None)
            markers = [0, 5, 6] if mark is None else 3 * [mark]
            for i, label in enumerate(["α ", "β ", "ɣ"]):
                fig.add_scatter(x=self.steps, y=[s.lattice.angles[i] for s in self.structures], mode='lines+markers',
                                name=label, marker_symbol=markers[i], row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = "αβɣ (degree)"+ "  "
            fig.layout['yaxis%u' % rcd.iax].tickformat = ".3r"

        elif what in ("alpha", "beta", "gamma"):
            i = ("alpha", "beta", "gamma").index(what)
            marker = kwargs.pop("marker", None)
            if marker is None:
                marker = {"alpha": 0, "beta": 5, "gamma": 6}[what]
            label = kwargs.pop("label", what)
            fig.add_scatter(x=self.steps, y=[s.lattice.angles[i] for s in self.structures], mode='lines+markers',
                            name=label, marker_symbol=marker, row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = r"%s (degree)" % latex_greek_2unicode(what)
            fig.layout['yaxis%u' % rcd.iax].tickformat = ".3r"

        elif what == "volume":
            marker = kwargs.pop("marker", 0)
            label = kwargs.pop("label", "Volume")
            fig.add_scatter(x=self.steps, y=[s.lattice.volume for s in self.structures], mode='lines+markers',
                            name=label, marker_symbol=marker, row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = 'V (A³)'

        elif what == "pressure":
            stress_cart_tensors, pressures = self.r.read_cart_stress_tensors()
            marker = kwargs.pop("marker", 0)
            label = kwargs.pop("label", "P")
            fig.add_scatter(x=self.steps, y=pressures, mode='lines+markers',
                            name=label, marker_symbol=marker, row=ply_row, col=ply_col, **kwargs)
            fig.layout['yaxis%u' % rcd.iax].title.text = 'P (GPa)'

        elif what == "forces":
            forces_hist = self.r.read_cart_forces()
            fmin_steps, fmax_steps, fmean_steps, fstd_steps = [], [], [], []
            for step in range(self.num_steps):
                forces = forces_hist[step]
                fmods = np.sqrt([np.dot(force, force) for force in forces])
                fmean_steps.append(fmods.mean())
                fstd_steps.append(fmods.std())
                fmin_steps.append(fmods.min())
                fmax_steps.append(fmods.max())

            mark = kwargs.pop("marker", None)
            markers = [0, 5, 6, 4] if mark is None else 4 * [mark]
            fig.add_scatter(x=self.steps, y=fmin_steps, mode='lines+markers',
                            name="min |F|", marker_symbol=markers[0], row=ply_row, col=ply_col, **kwargs)
            fig.add_scatter(x=self.steps, y=fmax_steps, mode='lines+markers',
                            name="max |F|", marker_symbol=markers[1], row=ply_row, col=ply_col, **kwargs)
            fig.add_scatter(x=self.steps, y=fmean_steps, mode='lines+markers',
                            name="mean |F|", marker_symbol=markers[2], row=ply_row, col=ply_col, **kwargs)
            fig.add_scatter(x=self.steps, y=fstd_steps, mode='lines+markers',
                            name="std |F|", marker_symbol=markers[3], row=ply_row, col=ply_col, **kwargs)
            label = "std |F|"
            fig.layout['yaxis%u' % rcd.iax].title.text = 'F stats (eV/A)'

        else:
            raise ValueError("Invalid value for what: `%s`" % str(what))

        fig.layout.legend.font.size = fontsize

    @add_fig_kwargs
    def plot(self, what_list=None, ax_list=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the evolution of structural parameters (lattice lengths, angles and volume)
        as well as pressure, info on forces and total energy with matplotlib.

        Args:
            what_list:
            ax_list: List of |matplotlib-Axes|. If None, a new figure is created.
            fontsize: fontsize for legend

        Returns: |matplotlib-Figure|
        """
        if what_list is None:
            what_list = ["abc", "angles", "volume", "pressure", "forces", "energy"]
        else:
            what_list = list_strings(what_list)

        nplots = len(what_list)
        nrows, ncols = 1, 1
        if nplots > 1:
            ncols = 2
            nrows = nplots // ncols + nplots % ncols

        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        # don't show the last ax if nplots is odd.
        if nplots % ncols != 0: ax_list[-1].axis("off")

        for what, ax in zip(what_list, ax_list):
            self.plot_ax(ax, what, fontsize=fontsize, marker="o")

        return fig

    @add_plotly_fig_kwargs
    def plotly(self, what_list=None, fig=None, fontsize=12, **kwargs):
        """
        Plot the evolution of structural parameters (lattice lengths, angles and volume)
        as well as pressure, info on forces and total energy with plotly.

        Args:
            what_list:
            fig: The fig for plot and the DOS plot. If None, a new figure is created.
            fontsize: fontsize for legend

        Returns: |plotly.graph_objects.Figure|
        """
        if what_list is None:
            what_list = ["abc", "angles", "volume", "pressure", "forces", "energy"]
        else:
            what_list = list_strings(what_list)

        nplots = len(what_list)
        nrows, ncols = 1, 1
        if nplots > 1:
            ncols = 2
            nrows = nplots // ncols + nplots % ncols

        if fig is None:
            fig, _ = get_figs_plotly(nrows=nrows, ncols=ncols, subplot_titles=[], sharex=True, sharey=False,
                                     vertical_spacing=0.05)

        for i, what in enumerate(what_list):
            rcd = PlotlyRowColDesc(i // ncols, i % ncols, nrows, ncols)
            self.plotly_traces(fig, what, rcd=rcd, fontsize=fontsize, marker=0)

        fig.layout['xaxis%u' % rcd.iax].title.text = 'Step'
        if nplots > 1:
            fig.layout['xaxis%s' % str(rcd.iax-1)].title.text = 'Step'

        return fig

    @add_plotly_fig_kwargs
    def plotly_energies(self, fig=None, fontsize=12, **kwargs):
        """
        Plot the total energies as function of the iteration step with plotly.
        """
        return self.plotly(what_list="energy", fig=fig, fontsize=fontsize, show=False, **kwargs)

    @add_fig_kwargs
    def plot_energies(self, ax=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the total energies as function of the iteration step with matplotlib.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        # TODO max force and pressure
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        terms = self.r.read_eterms()
        for key, values in terms.items():
            if np.all(values == 0.0): continue
            ax.plot(self.steps, values, marker="o", label=key)

        ax.set_xlabel('Step')
        ax.set_ylabel('Energies (eV)')
        ax.grid(True)
        ax.legend(loc='best', fontsize=fontsize, shadow=True)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)
        yield self.plot_energies(show=False)

    def yield_plotly_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plotly(show=False)
        yield self.plotly_energies(show=False)

    def mvplot_trajectories(self, colormap="hot", sampling=1, figure=None, show=True,
                            with_forces=True, **kwargs):  # pragma: no cover
        """
        Call mayavi_ to plot atomic trajectories and the variation of the unit cell.
        """
        from abipy.display import mvtk
        figure, mlab = mvtk.get_fig_mlab(figure=figure)
        style = "labels"
        line_width = 100
        mvtk.plot_structure(self.initial_structure, style=style, unit_cell_color=(1, 0, 0), figure=figure)
        mvtk.plot_structure(self.final_structure, style=style, unit_cell_color=(0, 0, 0), figure=figure)

        steps = np.arange(start=0, stop=self.num_steps, step=sampling)
        xcart_list = self.r.read_value("xcart") * units.bohr_to_ang
        for iatom in range(self.r.natom):
            x, y, z = xcart_list[::sampling, iatom, :].T
            #for i in zip(x, y, z): print(i)
            trajectory = mlab.plot3d(x, y, z, steps, colormap=colormap, tube_radius=None,
                                    line_width=line_width, figure=figure)
            mlab.colorbar(trajectory, title='Iteration', orientation='vertical')

        if with_forces:
            fcart_list = self.r.read_cart_forces(unit="eV ang^-1")
            for iatom in range(self.r.natom):
                x, y, z = xcart_list[::sampling, iatom, :].T
                u, v, w = fcart_list[::sampling, iatom, :].T
                q = mlab.quiver3d(x, y, z, u, v, w, figure=figure, colormap=colormap,
                                  line_width=line_width, scale_factor=10)
                #mlab.colorbar(q, title='Forces [eV/Ang]', orientation='vertical')

        if show: mlab.show()
        return figure

    def mvanimate(self, delay=500):  # pragma: no cover
        from abipy.display import mvtk
        figure, mlab = mvtk.get_fig_mlab(figure=None)
        style = "points"
        #mvtk.plot_structure(self.initial_structure, style=style, figure=figure)
        #mvtk.plot_structure(self.final_structure, style=style, figure=figure)

        xcart_list = self.r.read_value("xcart") * units.bohr_to_ang
        #t = np.arange(self.num_steps)
        #line_width = 2
        #for iatom in range(self.r.natom):
        #    x, y, z = xcart_list[:, iatom, :].T
        #    trajectory = mlab.plot3d(x, y, z, t, colormap=colormap, tube_radius=None, line_width=line_width, figure=figure)
        #mlab.colorbar(trajectory, title='Iteration', orientation='vertical')

        #x, y, z = xcart_list[0, :, :].T
        #nodes = mlab.points3d(x, y, z)
        #nodes.glyph.scale_mode = 'scale_by_vector'
        #this sets the vectors to be a 3x5000 vector showing some random scalars
        #nodes.mlab_source.dataset.point_data.vectors = np.tile( np.random.random((5000,)), (3,1))
        #nodes.mlab_source.dataset.point_data.scalars = np.random.random((5000,))

        @mlab.show
        @mlab.animate(delay=delay, ui=True)
        def anim():
            """Animate."""
            #for it in range(self.num_steps):
            for it, structure in enumerate(self.structures):
                print('Updating scene for iteration:', it)
                #mlab.clf(figure=figure)
                mvtk.plot_structure(structure, style=style, figure=figure)
                #x, y, z = xcart_list[it, :, :].T
                #nodes.mlab_source.set(x=x, y=y, z=z)
                #figure.scene.render()
                mlab.draw(figure=figure)
                yield

        anim()

    def get_panel(self, **kwargs):
        """
        Build panel with widgets to interact with the |HistFile| either in a notebook or in panel app.
        """
        from abipy.panels.hist import HistFilePanel
        return HistFilePanel(self).get_panel(**kwargs)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("hist = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(hist)"),
            nbv.new_code_cell("hist.plot_energies();"),
            nbv.new_code_cell("hist.plot();"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class HistRobot(Robot):
    """
    This robot analyzes the results contained in multiple HIST.nc_ files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: HistRobot
    """
    EXT = "HIST"

    def to_string(self, verbose: int = 0) -> str:
        """String representation with verbosity level ``verbose``."""
        s = ""
        if verbose:
            s = super().to_string(verbose=0)
        df = self.get_dataframe()
        s_df = "Table with final structures, pressures in GPa and force stats in eV/Ang:\n\n%s" % str(df)
        if s:
            return "\n".join([s, str(s_df)])
        else:
            return str(s_df)

    def get_dataframe(self, with_geo=True, index=None, abspath=False,
                      with_spglib=True, funcs=None, **kwargs) -> pd.DataFrame:
        """
        Return a |pandas-DataFrame| with the most important final results and the filenames as index.

        Args:
            with_geo: True if structure info should be added to the dataframe
            abspath: True if paths in index should be absolute. Default: Relative to getcwd().
            index: Index of the dataframe, if None, robot labels are used
            with_spglib: If True, spglib_ is invoked to get the space group symbol and number

        kwargs:
            attrs:
                List of additional attributes of the |GsrFile| to add to the |pandas-DataFrame|.
            funcs: Function or list of functions to execute to add more data to the DataFrame.
                Each function receives a |GsrFile| object and returns a tuple (key, value)
                where key is a string with the name of column and value is the value to be inserted.
        """
        # Add attributes specified by the users
        attrs = [
            "num_steps", "final_energy", "final_pressure",
            "final_fmin", "final_fmax", "final_fmean", "final_fstd", "final_drift",
            "initial_fmin", "initial_fmax", "initial_fmean", "initial_fstd", "initial_drift",
            # TODO add more columns but must update HIST file
            #"nsppol", "nspinor", "nspden",
            #"ecut", "pawecutdg", "tsmear", "nkpt",
        ] + kwargs.pop("attrs", [])

        rows, row_names = [], []
        for label, hist in self.items():
            row_names.append(label)
            d = OrderedDict()

            initial_fstas_dict = hist.get_fstats_dict(step=0)
            final_fstas_dict = hist.get_fstats_dict(step=-1)

            # Add info on structure.
            if with_geo:
                d.update(hist.final_structure.get_dict4pandas(with_spglib=with_spglib))

            for aname in attrs:
                if aname in ("final_fmin", "final_fmax", "final_fmean", "final_fstd", "final_drift",):
                    value = final_fstas_dict[aname.replace("final_", "")]
                elif aname in ("initial_fmin", "initial_fmax", "initial_fmean", "initial_fstd", "initial_drift"):
                    value = initial_fstas_dict[aname.replace("initial_", "")]
                else:
                    value = getattr(hist, aname, None)
                d[aname] = value

            # Execute functions
            if funcs is not None: d.update(self._exec_funcs(funcs, hist))
            rows.append(d)

        row_names = row_names if not abspath else self._to_relpaths(row_names)
        index = row_names if index is None else index
        return pd.DataFrame(rows, index=index, columns=list(rows[0].keys()))

    @property
    def what_list(self) -> list[str]:
        """List with all quantities that can be plotted (what_list)."""
        return ["energy", "abc", "angles", "volume", "pressure", "forces"]

    @add_fig_kwargs
    def gridplot(self, what_list=None, sharex="row", sharey="row", fontsize=8, **kwargs) -> Figure:
        """
        Plot the ``what`` value extracted from multiple HIST.nc_ files on a grid.

        Args:
            what_list: List of quantities to plot.
                Must be in ["energy", "abc", "angles", "volume", "pressure", "forces"]
            sharex: True if xaxis should be shared.
            sharey: True if yaxis should be shared.
            fontsize: fontsize for legend.

        Returns: |matplotlib-Figure|
        """
        what_list = list_strings(what_list) if what_list is not None else self.what_list

        # Build grid of plots.
        nrows, ncols = len(what_list), len(self)

        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=sharex, sharey=sharey, squeeze=False)
        ax_mat = np.reshape(ax_mat, (nrows, ncols))

        for irow, what in enumerate(what_list):
            for icol, hist in enumerate(self.abifiles):
                ax = ax_mat[irow, icol]
                ax.grid(True)
                hist.plot_ax(ax_mat[irow, icol], what, fontsize=fontsize, marker="o")

                if irow == 0:
                    ax.set_title(hist.relpath, fontsize=fontsize)
                if irow != nrows - 1:
                    set_visible(ax, False, "xlabel")
                if icol != 0:
                    set_visible(ax, False, "ylabel")

        return fig

    @add_fig_kwargs
    def combiplot(self, what_list=None, colormap="jet", fontsize=6, **kwargs) -> Figure:
        """
        Plot multiple HIST.nc_ files on a grid. One plot for each ``what`` value.

        Args:
            what_list: List of strings with the quantities to plot. If None, all quanties are plotted.
            colormap: matplotlib color map.
            fontsize: fontisize for legend.

        Returns: |matplotlib-Figure|.
        """
        what_list = (list_strings(what_list) if what_list is not None
            else ["energy", "a", "b", "c", "alpha", "beta", "gamma", "volume", "pressure"])

        num_plots, ncols, nrows = len(what_list), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()
        cmap = plt.get_cmap(colormap)

        for i, (ax, what) in enumerate(zip(ax_list, what_list)):
            for ih, hist in enumerate(self.abifiles):
                label = None if i != 0 else hist.relpath
                hist.plot_ax(ax, what, color=cmap(ih / len(self)), label=label, fontsize=fontsize)

            if label is not None:
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

            if i == len(ax_list) - 1:
                ax.set_xlabel("Step")
            else:
                ax.set_xlabel("")

        # Get around a bug in matplotlib.
        if num_plots % ncols != 0: ax_list[-1].axis('off')

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.gridplot(show=False)
        yield self.combiplot(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.HistRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("robot.get_dataframe()"),
            nbv.new_code_cell("for what in robot.what_list: robot.gridplot(what=what, tight_layout=True);"),
        ])

        # Mixins
        #nb.cells.extend(self.get_baserobot_code_cells())

        return self._write_nb_nbpath(nb, nbpath)


class HistReader(ETSF_Reader):
    """
    This object reads data from the HIST file.


    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: HistReader
    """

    @lazy_property
    def num_steps(self) -> int:
        """Number of iterations present in the HIST.nc_ file."""
        return self.read_dimvalue("time")

    @lazy_property
    def natom(self) -> int:
        """Number of atoms un the unit cell."""
        return self.read_dimvalue("natom")

    def read_all_structures(self) -> list[Structure]:
        """Return the list of structures at the different iteration steps."""
        rprimd_list = self.read_value("rprimd")
        xred_list = self.read_value("xred")

        # Alchemical mixing is not supported.
        num_pseudos = self.read_dimvalue("npsp")
        ntypat = self.read_dimvalue("ntypat")
        if num_pseudos != ntypat:
            raise NotImplementedError("Alchemical mixing is not supported, num_pseudos != ntypat")

        znucl, typat = self.read_value("znucl"), self.read_value("typat").astype(int)
        #print(znucl.dtype, typat)
        cart_forces_step = self.read_cart_forces(unit="eV ang^-1")

        structures = []
        #print("typat", type(typat))
        for step in range(self.num_steps):
            s = Structure.from_abivars(
                xred=xred_list[step],
                rprim=rprimd_list[step],
                acell=3 * [1.0],
                # FIXME ntypat, typat, znucl are missing!
                znucl=znucl,
                typat=typat,
            )
            s.add_site_property("cartesian_forces", cart_forces_step[step])
            structures.append(s)

        return structures

    def read_eterms(self, unit: str = "eV") -> AttrDict:
        """|AttrDict| with the decomposition of the total energy in units ``unit``"""
        return AttrDict(
            etotals=units.EnergyArray(self.read_value("etotal"), "Ha").to(unit),
            kinetic_terms=units.EnergyArray(self.read_value("ekin"), "Ha").to(unit),
            entropies=units.EnergyArray(self.read_value("entropy"), "Ha").to(unit),
        )

    def read_cart_forces(self, unit: str = "eV ang^-1") -> np.ndarray:
        """
        Read and return a |numpy-array| with the cartesian forces in unit ``unit``.
        Shape (num_steps, natom, 3)
        """
        return units.ArrayWithUnit(self.read_value("fcart"), "Ha bohr^-1").to(unit)

    def read_reduced_forces(self) -> np.ndarray:
        """
        Read and return a |numpy-array| with the forces in reduced coordinates
        Shape (num_steps, natom, 3)
        """
        return self.read_value("fred")

    def read_cart_stress_tensors(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Return the stress tensors (nstep x 3 x 3) in cartesian coordinates (GPa)
        and the list of pressures in GPa unit.
        """
        # Abinit stores 6 unique components of this symmetric 3x3 tensor:
        # Given in order (1,1), (2,2), (3,3), (3,2), (3,1), (2,1).
        c = self.read_value("strten")
        tensors = np.empty((self.num_steps, 3, 3), dtype=float)

        for step in range(self.num_steps):
            for i in range(3): tensors[step, i, i] = c[step, i]
            for p, (i, j) in enumerate(((2, 1), (2, 0), (1, 0))):
                tensors[step, i, j] = c[step, 3+p]
                tensors[step, j, i] = c[step, 3+p]

        tensors *= abu.HaBohr3_GPa
        pressures = np.empty(self.num_steps)
        for step, tensor in enumerate(tensors):
            pressures[step] = - tensor.trace() / 3

        return tensors, pressures



"""
Integration tests for flows (require pytest, ABINIT and a properly configured environment)
"""
from __future__ import annotations

import os
import unittest
#import numpy.testing.utils as nptu
import numpy.testing as nptu
import abipy.data as abidata
import abipy.flowtk as flowtk
import abipy.flowtk.abiphonopy as abiph

from abipy.abio.factories import gs_input
from abipy.core.testing import has_phonopy


def itest_phonopy_flow(fwp, tvars):
    """
    Testing phonopy flow with the scheduler.
    """
    if not has_phonopy():
        raise unittest.SkipTest("This test requires phonopy")

    #print("tvars:\n %s" % str(tvars))
    si_structure = abidata.structure_from_cif("si.cif")

    gsinp = gs_input(si_structure, pseudos=abidata.pseudos("14si.pspnc"), kppa=10, ecut=2, spin_mode="unpolarized")
    gsinp["paral_kgb"] = tvars.paral_kgb

    flow = flowtk.Flow(workdir=fwp.workdir, manager=fwp.manager)
    scdims = [2, 2, 2]
    phpy_work = abiph.PhonopyWork.from_gs_input(gsinp, scdims=scdims,
                                                phonopy_kwargs=None, displ_kwargs=None)
    flow.register_work(phpy_work)

    assert hasattr(phpy_work, "phonon")
    assert len(phpy_work.phonopy_tasks) == len(phpy_work)
    assert len(phpy_work.phonopy_tasks) == 1
    assert len(phpy_work.bec_tasks) == 0
    nptu.assert_equal(scdims, phpy_work.scdims)

    # Will remove output files (WFK)
    flow.set_garbage_collector()
    flow.use_smartio()

    flow.build_and_pickle_dump(abivalidate=True)
    scheduler = flow.make_scheduler()
    assert scheduler.start() == 0

    assert not scheduler.exceptions
    flow.show_status()
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()
    assert all(work.finalized for work in flow)

    # The WFK files should have been removed because we called set_garbage_collector
    for task in flow[0]:
        assert not task.outdir.has_abiext("WFK")

    out_filenames = set([os.path.basename(f) for f in phpy_work.outdir.list_filepaths()])
    nmiss = 0
    for f in ["POSCAR", "disp.yaml", "FORCE_SETS", "band.conf", "dos.conf", "band-dos.conf", "README.md"]:
        if f not in out_filenames:
            nmiss += 1
            print("Cannot find %s in work.outdir" % f)
    assert nmiss == 0


def itest_phonopy_gruneisen_flow(fwp, tvars):
    """
    Testing phonopy Gruneisen flow with the scheduler.
    """
    if not has_phonopy():
        raise unittest.SkipTest("This test requires phonopy")

    #print("tvars:\n %s" % str(tvars))
    si_structure = abidata.structure_from_cif("si.cif")

    gsinp = gs_input(si_structure, pseudos=abidata.pseudos("14si.pspnc"), kppa=10, ecut=2, spin_mode="unpolarized")
    gsinp["paral_kgb"] = tvars.paral_kgb

    flow = flowtk.Flow(workdir=fwp.workdir, manager=fwp.manager)
    scdims = [2, 2, 2]

    # Grunesein with phonopy
    grun_work = abiph.PhonopyGruneisenWork.from_gs_input(gsinp, voldelta=0.1, scdims=scdims,
                                                         phonopy_kwargs=None, displ_kwargs=None)
    flow.register_work(grun_work)
    assert len(grun_work) == 3
    nptu.assert_equal(scdims, grun_work.scdims)

    # Will remove output files (WFK)
    flow.set_garbage_collector()
    flow.use_smartio()

    flow.build_and_pickle_dump(abivalidate=True)
    scheduler = flow.make_scheduler()
    assert scheduler.start() == 0

    assert not scheduler.exceptions
    flow.show_status()
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()
    # Initialial work + 3 phonopy works.
    assert len(flow) == 4
    assert all(work.finalized for work in flow)

    # The WFK files should have been removed because we called set_garbage_collector
    # FIXME: This does not work because new works that have been created.
    #for task in flow.iflat_tasks():
    #    assert not task.outdir.has_abiext("WFK")

    for work in flow[1:]:
        out_filenames = set([os.path.basename(f) for f in work.outdir.list_filepaths()])
        nmiss = 0
        for f in ["POSCAR", "disp.yaml", "FORCE_SETS", "band.conf", "dos.conf", "band-dos.conf", "README.md"]:
            if f not in out_filenames:
                nmiss += 1
                print("Cannot find %s in work.outdir" % f)
        assert nmiss == 0



#!/usr/bin/env python
r"""
ElectronDosPlotter
==================

This example shows how to plot several Electron Dos on a grid
using the results stored in the ni_666k_GSR.nc file.
"""

#%%
# We start by defining a list with the paths to the GSR.nc files
# In this case, for simplicity, we use the same file but we must
# use different labels when adding them to the plotter with the add_edos method.

from abipy import abilab
import abipy.data as abidata

edos_paths = [abidata.ref_file("ni_666k_GSR.nc")]

plotter = abilab.ElectronDosPlotter()
plotter.add_edos("Ni", edos_paths[0])
plotter.add_edos("Same Ni", edos_paths[0])

#%%
# To visualize the results on a grid with matplotlib, use:
plotter.gridplot(title="Electron DOS on grid")

#%%
# Plotly version:
plotter.gridplotly(title="Electron DOS on grid")




#!/usr/bin/env python
"""Integration tests for the scheduler."""
from __future__ import annotations

import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk

from abipy.flowtk import mocks


def make_scf_nscf_inputs(paral_kgb=1):
    """Returns two input files: GS run and NSCF on a high symmetry k-mesh."""
    multi = abilab.MultiDataset(structure=abidata.cif_file("si.cif"),
                              pseudos=abidata.pseudos("14si.pspnc"), ndtset=2)

    # Global variables
    ecut = 4
    global_vars = dict(ecut=ecut, nband=8, nstep=15, paral_kgb=paral_kgb)

    if multi.ispaw:
        global_vars.update(pawecutdg=2 * ecut)

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(ngkpt=[2, 2, 2], shiftk=[0, 0, 0])
    multi[0].set_vars(tolvrs=1e-4)

    # Dataset 2 (NSCF run)
    kptbounds = [
        [0.5, 0.0, 0.0], # L point
        [0.0, 0.0, 0.0], # Gamma point
        [0.0, 0.5, 0.5], # X point
    ]

    multi[1].set_kpath(ndivsm=2, kptbounds=kptbounds)
    multi[1].set_vars(tolwfr=1e-4)

    # Generate two input files for the GS and the NSCF run
    scf_input, nscf_input = multi.split_datasets()
    return scf_input, nscf_input


def itest_flow_with_deadlocks(fwp):
    """
    Test the behaviour of the scheduler in the presence of a deadlock
    when we ignore errored tasks and we try to run all tasks in the flow.
    The scheduler should detect the deadlock and exit when no other task can be executed.
    """
    # Get the SCF and the NSCF input.
    scf_input, nscf_input = make_scf_nscf_inputs()

    # Build the flow.
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)
    work0 = flowtk.BandStructureWork(scf_input, nscf_input, dos_inputs=nscf_input)
    flow.register_work(work0)
    scf_task, nscf_task, dos_task = work0[0], work0[1], work0[2]

    work1 = flowtk.Work()
    work1.register_nscf_task(nscf_input, deps={scf_task: "DEN", dos_task: "WFK"})
    # This task will deadlock when nscf_task reaches S_ERROR.
    work1.register_nscf_task(nscf_input, deps={scf_task: "DEN", nscf_task: "WFK"})
    flow.register_work(work1)

    flow.allocate()

    # Mock an Errored nscf_task. This will cause a deadlock in the flow.
    nscf_task = mocks.change_task_start(nscf_task)

    # Here we set max_num_abierrs to a very large number.
    sched = flow.make_scheduler()
    sched.max_num_abierrs = 10000
    assert sched.start() == 0
    flow.check_status(show=True)

    assert not flow.all_ok
    assert all(task.status == task.S_OK for task in [scf_task, dos_task, work1[0]])
    assert all(task.status == task.S_ERROR for task in [nscf_task])
    g = flow.find_deadlocks()
    assert g.deadlocked and not g.runnables and not g.running
    assert work1[1] in g.deadlocked


def itest_flow_without_runnable_tasks(fwp):
    """
    Test the behaviour of the scheduler when we ignore errors and
    all the task that can be executed have been submitted.
    The scheduler should detect this condition and exit.
    """
    # Get the SCF and the NSCF input.
    scf_input, nscf_input = make_scf_nscf_inputs()

    # Build the flow.
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)
    work0 = flowtk.BandStructureWork(scf_input, nscf_input)
    flow.register_work(work0)
    scf_task, nscf_task = work0.scf_task, work0.nscf_task

    flow.allocate()

    # Mock an Errored nscf_task. This will cause a deadlock in the flow.
    nscf_task = mocks.change_task_start(nscf_task)

    sched = flow.make_scheduler()
    sched.max_num_abierrs = 10000
    assert sched.start() == 0
    flow.check_status(show=True)

    assert not flow.all_ok
    assert scf_task.status == scf_task.S_OK
    assert nscf_task.status == nscf_task.S_ERROR

    g = flow.find_deadlocks()
    assert not g.deadlocked and not g.runnables and not g.running



""""Panels for HIST files."""
from __future__ import annotations

#import param
import panel as pn
import panel.widgets as pnw
import bokeh.models.widgets as bkw

from abipy.dynamics.hist import HistFile
from abipy.panels.core import AbipyParameterized, mpl, ply, depends_on_btn_click


class HistFilePanel(AbipyParameterized):
    """
    Panel with widgets to interact with a |HistFile|.
    """

    def __init__(self, hist: HistFile, **params):
        self.hist = hist

        _what_list = ["abc", "angles", "energy", "volume", "pressure", "forces"]
        self.what_list = pnw.CheckBoxGroup(name="Select", value=_what_list, options=_what_list, inline=False)
        self.plot_relax_btn = pnw.Button(name="Plot relaxation", button_type="primary")

        self.appname = pnw.Select(name="Viewer", value="ovito", options=["ovito", "mayavi", "vtk"])
        self.to_unit_cell = pnw.Checkbox(name="To unit cell")
        self.view_relax_btn = pnw.Button(name="View relaxation", button_type="primary")

        super().__init__(**params)

    @depends_on_btn_click('plot_relax_btn')
    def on_plot_relax_btn(self) -> pn.Column:
        """
        Plot the evolution of structural parameters (lattice lengths, angles and volume)
        as well as pressure, info on forces and total energy.
        """
        col = pn.Column(sizing_mode="stretch_width"); ca = col.append
        for what in self.what_list.value:
            #ca(f"## {what}")
            ca(ply(self.hist.plotly(what, title=what, show=False)))

        return col

    @depends_on_btn_click('view_relax_btn')
    def on_view_relax_btn(self):
        """
        Visalize the structural relaxation with an external application.
        """
        return self.hist.visualize(appname=self.appname.value, to_unit_cell=self.to_unit_cell.value)

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the HIST.nc file."""
        d = {}

        d["Summary"] = self.get_summary_view_for_abiobj(self.hist)
        d["Plot"] = pn.Row(
                self.pws_col(["## Plot Options", "what_list", "plot_relax_btn"]),
                self.on_plot_relax_btn
        )

        if not self.has_remote_server:
            # As we don't have visualizers that can work in remote server mode,
            # this tab should not be created.
            d["Visualize"] = pn.Row(
                    pn.Column(self.appname, self.to_unit_cell, self.view_relax_btn),
                    self.on_view_relax_btn
            )

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))



# coding: utf-8
"""The interatomic force constants calculated by anaddb."""
from __future__ import annotations

import numpy as np

from typing import Union
from monty.functools import lazy_property
from abipy.core.structure import Structure
from abipy.core.mixins import Has_Structure
from abipy.tools.typing import Figure
from abipy.iotools import ETSF_Reader
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt


class InteratomicForceConstants(Has_Structure):
    """
    The interatomic force constants calculated by anaddb.
    Read from the anaddb.nc file.
    """

    def __init__(self, structure, atoms_indices, neighbours_indices, ifc_cart_coord,
                 ifc_cart_coord_short_range, local_vectors, distances, atoms_cart_coord,
                 ifc_weights):
        """
        Args:
            structure: |Structure| object.
            atoms_indices: List of integers representing the indices in the structure of the analyzed atoms.
            neighbours_indices: List of integers representing the indices in the structure of the neighbour atoms.
            ifc_cart_coord: ifc in Cartesian coordinates
            ifc_cart_coord_short_range: short range part of the ifc in Cartesian coordinates
            local_vectors: local basis used to determine the ifc_local_coord
            atoms_cart_coord: cartesian coordinates in Bohr of the atoms corresponding to the atoms in
                atoms_index
            ifc_weights: list of weights used for the interpolation for each atoms in atoms_index
        """
        self._structure = structure
        self.atoms_indices = atoms_indices
        self.neighbours_indices = neighbours_indices
        self.ifc_cart_coord = ifc_cart_coord
        self.ifc_cart_coord_short_range = ifc_cart_coord_short_range
        self.local_vectors = local_vectors
        self.distances = distances
        self.atoms_cart_coord = atoms_cart_coord
        self.ifc_weights = ifc_weights

    @property
    def number_of_atoms(self) -> int:
        """Number of atoms in the structure."""
        return len(self.structure)

    @classmethod
    def from_file(cls, filepath: str) -> InteratomicForceConstants:
        """Create the object from a netcdf file."""
        #
        # Netcdf arrays on disk (NB: Fortran conventions)
        #
        # nctkarr_t('ifc_atoms_indices', "i", "natifc"),&
        # nctkarr_t('ifc_neighbours_indices', "i", "ifcout, natifc"),&
        # nctkarr_t('ifc_distances', "dp", "ifcout, natifc "),&
        # nctkarr_t('ifc_matrix_cart_coord', "dp", "three, three, ifcout, natifc"),&
        # nctkarr_t('ifc_atoms_cart_coord', "dp", "three, ifcout, natifc"),&
        # nctkarr_t('ifc_weights', "dp", "ifcout, natifc")])
        # if (Ifc%dipdip==1) then
        #   nctkarr_t('ifc_matrix_cart_coord_short_range', "dp", "three, three, ifcout, natifc")])
        # if (ifcana==1) then
        #   nctkarr_t('ifc_local_vectors', "dp", "three, three, ifcout, natifc")])

        with ETSF_Reader(filepath) as r:
            try:
                structure = r.read_structure()
                atoms_indices = r.read_value("ifc_atoms_indices") - 1
                neighbours_indices = r.read_value("ifc_neighbours_indices") - 1
                distances = r.read_value("ifc_distances")
                ifc_cart_coord = r.read_value("ifc_matrix_cart_coord")
                ifc_cart_coord_short_range = r.read_value("ifc_matrix_cart_coord_short_range", default=None)
                local_vectors = r.read_value("ifc_local_vectors", default=None)
                # The following are set to None by default to keep backward
                # compatibility with anaddb.nc files generated by versions of abinit<9.2
                atoms_cart_coord = r.read_value("ifc_atoms_cart_coord", default=None)
                ifc_weights = r.read_value("ifc_weights", default=None)
            except Exception:
                import traceback
                msg = traceback.format_exc()
                msg += ("Error while trying to read IFCs from file.\n"
                        "Verify that the required variables are used in anaddb: ifcflag, natifc, atifc, ifcout\n")
                raise ValueError(msg)

            return cls(structure=structure, atoms_indices=atoms_indices, neighbours_indices=neighbours_indices,
                       ifc_cart_coord=ifc_cart_coord,
                       ifc_cart_coord_short_range=ifc_cart_coord_short_range, local_vectors=local_vectors,
                       distances=distances, atoms_cart_coord=atoms_cart_coord, ifc_weights=ifc_weights)

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self._structure

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []; app = lines.append
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")

        return "\n".join(lines)

    @property
    def number_of_neighbours(self) -> int:
        """Number of neighbouring atoms for which the ifc are present. ifcout in anaddb."""
        return np.shape(self.neighbours_indices)[1]

    @lazy_property
    def ifc_cart_coord_ewald(self) -> Union[None, np.ndarray]:
        """Ewald part of the IFCs in cartesian coordinates."""
        if self.ifc_cart_coord_short_range is None:
            return None
        else:
            return self.ifc_cart_coord - self.ifc_cart_coord_short_range

    @lazy_property
    def ifc_local_coord(self) -> Union[None, np.ndarray]:
        """IFCs in local coordinates."""
        if self.local_vectors is None:
            return None
        else:
            return np.einsum("ktli,ktij,ktuj->ktlu", self.local_vectors, self.ifc_cart_coord, self.local_vectors)

    @lazy_property
    def ifc_local_coord_short_range(self) -> Union[None, np.ndarray]:
        """Short range part of the IFCs in cartesian coordinates."""
        if self.local_vectors is None:
            return None
        else:
            return np.einsum("ktli,ktij,ktuj->ktlu", self.local_vectors, self.ifc_cart_coord_short_range, self.local_vectors)

    @lazy_property
    def ifc_local_coord_ewald(self) -> np.ndarray:
        """Ewald part of the IFCs in local coordinates."""
        return np.einsum("ktli,ktij,ktuj->ktlu", self.local_vectors, self.ifc_cart_coord_ewald, self.local_vectors)

    def _filter_ifc_indices(self, atom_indices=None, atom_element=None, neighbour_element=None, min_dist=None, max_dist=None):
        """
        Internal method that provides the indices of the neighouring atoms in self.neighbours_indices that satisfy
        the required conditions. All the arguments are optional. If None the filter will not be applied.

        Args:
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
        """

        if atom_indices is not None and atom_element is not None:
            raise ValueError("atom_index and atom_element cannot be specified simultaneously")

        if atom_indices is not None and not isinstance(atom_indices, (list, tuple)):
            atom_indices = [atom_indices]

        if atom_element:
            atom_indices = self.structure.indices_from_symbol(atom_element)

        if atom_indices is None:
            atom_indices = range(len(self.structure))

        # apply the filter: construct matrices of num_atoms * num_neighbours size, all conditions should be satisfied.
        ind = np.where(
            (np.tile(np.in1d(self.atoms_indices, atom_indices), [self.number_of_neighbours, 1])).T &
            (self.distances > min_dist if min_dist is not None else True) &
            (self.distances < max_dist if max_dist is not None else True) &
            (np.in1d(self.neighbours_indices, self.structure.indices_from_symbol(neighbour_element))
             .reshape(self.number_of_atoms, self.number_of_neighbours) if neighbour_element is not None else True)
        )

        return ind

    def get_ifc_cartesian(self, atom_indices=None, atom_element=None, neighbour_element=None, min_dist=None, max_dist=None):
        """
        Filters the IFCs in cartesian coordinates.
        All the arguments are optional. If None the filter will not be applied.
        Returns two arrays containing the distances and the corresponding filtered IFCs.

        Args:
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
        """
        ind = self._filter_ifc_indices(atom_indices=atom_indices, atom_element=atom_element,
                                       neighbour_element=neighbour_element, min_dist=min_dist, max_dist=max_dist)

        return self.distances[ind], self.ifc_cart_coord[ind]

    def get_ifc_local(self, atom_indices=None, atom_element=None, neighbour_element=None, min_dist=None, max_dist=None):
        """
        Filters the IFCs in local coordinates
        All the arguments are optional. If None the filter will not be applied.
        Returns two arrays containing the distances and the corresponding filtered IFCs.

        Args:
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
        """
        if self.local_vectors is None:
            raise ValueError("Local coordinates are missing. Run anaddb with ifcana = 1")

        ind = self._filter_ifc_indices(atom_indices=atom_indices, atom_element=atom_element,
                                       neighbour_element=neighbour_element, min_dist=min_dist, max_dist=max_dist)

        return self.distances[ind], self.ifc_local_coord[ind]

    def get_plot_ifc(self, ifc, atom_indices=None, atom_element=None, neighbour_element=None, min_dist=None,
                     max_dist=None, yscale="log", ax=None, **kwargs) -> Figure:
        """
        Plots the specified IFCs, filtered according to the optional arguments.
        An array with shape number_of_atoms * number_of_neighbours, so only one
        of the components of the IFC matrix can be plotted at a time.

        Args:
            ifc: an array with shape number_of_atoms * number_of_neighbours of the ifc that should be plotted
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
            yscale: The axis scale type to apply: {"linear", "log", "symlog", "logit"}
            ax: |matplotlib-Axes| or None if a new figure should be created.
            kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to blue, symbol to 'o' and lw to 0

        Returns: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        ind = self._filter_ifc_indices(atom_indices=atom_indices, atom_element=atom_element,
                                       neighbour_element=neighbour_element, min_dist=min_dist, max_dist=max_dist)

        dist, filtered_ifc = self.distances[ind], ifc[ind]

        if 'color' not in kwargs: kwargs['color'] = 'blue'
        if 'marker' not in kwargs: kwargs['marker'] = 'o'
        if 'linewidth' not in kwargs and 'lw' not in kwargs: kwargs['lw'] = 0

        ax.set_yscale(yscale)
        ylabel = r'IFC (Ha/Bohr$^2$)'
        if yscale in ("log", "symlog", "logit"):
            filtered_ifc = np.abs(filtered_ifc)
            if yscale == "logit": filtered_ifc /= filtered_ifc.max()
            ylabel = f"{ylabel} ({yscale} scale)"

        ax.set_xlabel('Distance (Bohr)')
        ax.set_ylabel(ylabel)
        ax.grid(True)

        ax.plot(dist, filtered_ifc, **kwargs)

        return fig

    @add_fig_kwargs
    def plot_longitudinal_ifc(self, atom_indices=None, atom_element=None, neighbour_element=None, min_dist=None,
                              max_dist=None, yscale="log", ax=None, **kwargs) -> Figure:
        """
        Plots the total longitudinal IFCs in local coordinates, filtered according to the optional arguments.

        Args:
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
            yscale: The axis scale type to apply: {"linear", "log", "symlog", "logit"}
                See <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_yscale.html>
            ax: |matplotlib-Axes| or None if a new figure should be created.
            kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to blue, symbol to 'o' and lw to 0

        Returns: |matplotlib-Figure|
        """
        if self.local_vectors is None:
            raise ValueError("Local coordinates are missing. Run anaddb with ifcana = 1")

        return self.get_plot_ifc(self.ifc_local_coord[:, :, 0, 0], atom_indices=atom_indices, atom_element=atom_element,
                                 neighbour_element=neighbour_element, min_dist=min_dist, max_dist=max_dist,
                                 yscale=yscale, ax=ax, **kwargs)

    @add_fig_kwargs
    def plot_longitudinal_ifc_short_range(self, atom_indices=None, atom_element=None, neighbour_element=None,
                                          min_dist=None, max_dist=None, yscale="log", ax=None, **kwargs) -> Figure:
        """
        Plots the short range longitudinal IFCs in local coordinates, filtered according to the optional arguments.

        Args:
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
            yscale: The axis scale type to apply: {"linear", "log", "symlog", "logit"}
                See <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_yscale.html>
            ax: |matplotlib-Axes| or None if a new figure should be created.
            kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to blue, symbol to 'o' and lw to 0

        Returns: |matplotlib-Figure|
        """
        if self.local_vectors is None:
            raise ValueError("Local coordinates are missing. Run anaddb with ifcana = 1")

        if self.ifc_local_coord_short_range is None:
            raise ValueError("Ewald contribution is missing, Run anaddb with dipdip = 1")

        return self.get_plot_ifc(self.ifc_local_coord_short_range[:, :, 0, 0], atom_indices=atom_indices,
                                 atom_element=atom_element, neighbour_element=neighbour_element, min_dist=min_dist,
                                 max_dist=max_dist, yscale=yscale, ax=ax, **kwargs)

    @add_fig_kwargs
    def plot_longitudinal_ifc_ewald(self, atom_indices=None, atom_element=None, neighbour_element=None,
                                    min_dist=None, max_dist=None, yscale="log", ax=None, **kwargs) -> Figure:
        """
        Plots the Ewald part of the IFCs in local coordinates, filtered according to the optional arguments.

        Args:
            atom_indices: a list of atom indices in the structure. Only neighbours of these atoms will be considered.
            atom_element: symbol of an element in the structure. Only neighbours of these atoms will be considered.
            neighbour_element: symbol of an element in the structure. Only neighbours of this specie will be considered.
            min_dist: minimum distance between atoms and neighbours.
            max_dist: maximum distance between atoms and neighbours.
            yscale: The axis scale type to apply: {"linear", "log", "symlog", "logit"}
                See <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.set_yscale.html>
            ax: |matplotlib-Axes| or None if a new figure should be created.
            kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to blue, symbol to 'o' and lw to 0

        Returns: |matplotlib-Figure|
        """
        if self.local_vectors is None:
            raise ValueError("Local coordinates are missing. Run anaddb with ifcana = 1")

        if self.ifc_local_coord_ewald is None:
            raise ValueError("Ewald contribution is missing, Run anaddb with dipdip = 1")

        return self.get_plot_ifc(self.ifc_local_coord_ewald[:, :, 0, 0], atom_indices=atom_indices,
                                 atom_element=atom_element, neighbour_element=neighbour_element, min_dist=min_dist,
                                 max_dist=max_dist, yscale=yscale, ax=ax, **kwargs)



#!/usr/bin/env python
r"""
self-consistent GW
==================

This example shows how to visualize the SCGW QP amplitudes in the KS basis set.
"""
import abipy.data as abidata
from abipy.abilab import abiopen

sigres = abiopen(abidata.ref_file("QPSC_SIGRES.nc"))

#print("calctyp",sigres.gwcalctyp)
#sigres.print_qps()

#qp = sigres.get_qpcorr(spin=0, kpoint=(0,0,0), band=0)
#print(qp)

# Visualize <KS_b|QP_b'> matrix at the Gamma point
sigres.plot_eigvec_qp(kpoint=[0, 0, 0])

# Plot <KS_b|QP_b'> components for all k-points in the SIGRES.
sigres.plot_eigvec_qp()



# coding: utf-8
"""Tools to build ipython widgets."""
from __future__ import annotations

import ipywidgets as ipw

from collections import OrderedDict


#def add_docstrings(*tuples):
#    """
#    This decorator adds to the docstring the documentation for functions.
#    When writing high-level API, it's quite common to call thirdy-party functions
#    with a restricted set of arguments while optional keyword arguments are
#    collected in an optional dictionary.
#
#    The first item of the tuple contains the function (python object) wrapped by the code.
#    The second item is list of strings with the name of the actual arguments passed to function.
#    """
#    from functools import wraps
#    def wrapper(func):
#        @wraps(func)
#        def wrapped_func(*args, **kwargs):
#            return func(*args, **kwargs)
#
#        # Add docstrings for the functions that will be called by func.
#        lines = []
#        app = lines.append
#        for t in tuples:
#            fname = t[0].__name__
#            # List of strings or string.
#            if isinstance(t[1], (list, tuple)):
#                fargs = ",".join("`%s`" % a for a in t[1])
#            else:
#                fargs = "`%s`" % t[1]
#            app("\n%s are passed to function :func:`%s` in module :mod:`%s`" % (fargs, fname, t[0].__module__))
#            app("Docstring of `%s`:" % fname)
#            app(t[0].__doc__)
#        s = "\n".join(lines)
#
#        if wrapped_func.__doc__ is not None:
#            # Add s at the end of the docstring.
#            wrapped_func.__doc__ += "\n" + s
#        else:
#            # Use s
#            wrapped_func.__doc__ = s
#        return wrapped_func
#    return wrapper


def widget2py(*args):
    l = [None if a == "None" else a for a in args]
    return l[0] if len(l) == 1 else l


def str2bool_or_none(*args):
    d = {"None": None, "True": True, "False": False}
    l = [d[a] for a in args]
    return l[0] if len(l) == 1 else l


# Taken from matplotlib.markers.MarkerStyle (replaced dict with OrderedDict).
_mpl_markers = OrderedDict([
    ('.', 'point'),
    (',', 'pixel'),
    ('o', 'circle'),
    ('v', 'triangle_down'),
    ('^', 'triangle_up'),
    ('<', 'triangle_left'),
    ('>', 'triangle_right'),
    ('1', 'tri_down'),
    ('2', 'tri_up'),
    ('3', 'tri_left'),
    ('4', 'tri_right'),
    ('8', 'octagon'),
    ('s', 'square'),
    ('p', 'pentagon'),
    ('*', 'star'),
    ('h', 'hexagon1'),
    ('H', 'hexagon2'),
    ('+', 'plus'),
    ('x', 'x'),
    ('D', 'diamond'),
    ('d', 'thin_diamond'),
    ('|', 'vline'),
    ('_', 'hline'),
    #(TICKLEFT: 'tickleft',
    #(TICKRIGHT: 'tickright',
    #(TICKUP: 'tickup',
    #(TICKDOWN: 'tickdown',
    #(CARETLEFT: 'caretleft',
    #(CARETRIGHT: 'caretright',
    #(CARETUP: 'caretup',
    #(CARETDOWN: 'caretdown',
    ("None", 'nothing'),
    (None, 'nothing'),
    (' ', 'nothing'),
    ('', 'nothing'),
])


def markers_dropdown(default="o"):
    return ipw.Dropdown(
        options={name: key for key, name in _mpl_markers.items()},
        value=default,
        description='marker',
    )


_mpl_colors = OrderedDict([
    ("None", "None"),
    ("blue", "b"),
    ("green", "g"),
    ("red", "r"),
    ("cyan", "c"),
    ("magenta", "m"),
    ("yellow", "y"),
    ("black", "k"),
    ("white", "w"),
])


def colors_dropdow(default="None"):
    return ipw.Dropdown(
        options=_mpl_colors,
        value=default,
        description='color',
    )


def linewidth_slider(default=1, orientation="horizontal"):
    return ipw.FloatSlider(
        value=default,
        min=0,
        max=10,
        step=0.5,
        description='linewidth',
        orientation=orientation,
        readout_format='.1f'
    )


def size_slider(default=5, orientation="horizontal"):
    return ipw.FloatSlider(
        value=default,
        min=0,
        max=20,
        step=0.5,
        description='size',
        orientation=orientation,
        readout_format='.1f'
    )


def saturation_slider(default=0.75, orientation="horizontal"):
    return ipw.FloatSlider(
        value=default,
        min=0,
        max=1,
        step=0.05,
        description='saturation',
        orientation=orientation,
        readout_format='.1f'
    )


# Have colormaps separated into categories:
# http://matplotlib.org/examples/color/colormaps_reference.html
_mpl_categ_cmaps = OrderedDict([
    #('Perceptually Uniform Sequential',
    ('Uniform',        ['viridis', 'inferno', 'plasma', 'magma']),
    ('Sequential',     ['Blues', 'BuGn', 'BuPu',
                        'GnBu', 'Greens', 'Greys', 'Oranges', 'OrRd',
                        'PuBu', 'PuBuGn', 'PuRd', 'Purples', 'RdPu',
                        'Reds', 'YlGn', 'YlGnBu', 'YlOrBr', 'YlOrRd']),
    ('Sequential(2)',  ['afmhot', 'autumn', 'bone', 'cool',
                        'copper', 'gist_heat', 'gray', 'hot',
                        'pink', 'spring', 'summer', 'winter']),
    ('Diverging',      ['BrBG', 'bwr', 'coolwarm', 'PiYG', 'PRGn', 'PuOr',
                        'RdBu', 'RdGy', 'RdYlBu', 'RdYlGn', 'Spectral',
                        'seismic']),
    ('Qualitative',    ['Accent', 'Dark2', 'Paired', 'Pastel1',
                        'Pastel2', 'Set1', 'Set2', 'Set3']),
    ('Miscellaneous',  ['gist_earth', 'terrain', 'ocean', 'gist_stern',
                        'brg', 'CMRmap', 'cubehelix',
                        'gnuplot', 'gnuplot2', 'gist_ncar',
                        'nipy_spectral', 'jet', 'rainbow',
                        'gist_rainbow', 'hsv', 'flag', 'prism'])
])

# flat list.
_mpl_cmaps = [cm for sublist in _mpl_categ_cmaps.values() for cm in sublist]


def colormap_widget(default=None):
    options = _mpl_cmaps
    value = options[0]
    if default is not None:
        value = default
        if default not in _mpl_cmaps: options[:].insert(0, value)
    return ipw.Dropdown(options=options, value=value, description='colormap')


#def colormap_widget():
#    from IPython.display import display, clear_output
#    w_type = ipw.Dropdown(options=list(_mpl_categ_cmaps.keys()), description='colormap category')
#    w_cmap = ipw.Dropdown(options=_mpl_categ_cmaps["Uniform"], description='colormap name')
#
#    def on_value_change(change):
#        print(change)
#        print(w_cmap.value)
#        w_cmap.options = _mpl_categ_cmaps[w_type.value]
#        print(w_cmap.value)
#
#    w_type.observe(on_value_change, names='value')
#    box = ipw.HBox(children=[w_type, w_cmap])
#    return display(box)



#!/usr/bin/env python
from __future__ import print_function, unicode_literals, division, absolute_import
import sys

from abipy.data import AbinitFilesGenerator


class MyGenerator(AbinitFilesGenerator):
    """This class generates the output files used in the unit tests and in the examples."""
    # Subclasses must define the following class attributes:
    # List of pseudos (basenames in abipy/data/pseudos)
    pseudos = ["13al.981214.fhi"]

    # Mapping old_name --> new_name for the output files that must be preserved.
    files_to_save = {
        "out_DS4_SIGRES.nc": "al_g0w0_sigmaw_SIGRES.nc",
    }


if __name__ == "__main__":
    sys.exit(MyGenerator().run())



#!/usr/bin/env python
r"""
MgB2 fatbands
=============

This example shows how to plot the L-projected fatbands of MgB2
using the FATBANDS.nc files produced by abinit with prtdos 3.
See also PhysRevLett.86.4656
"""

#%%
# Open the file (alternatively one can use the shell and `abiopen.py FILE -nb`
# to open the file in a jupyter notebook)
# Note that this file has been produced on a k-path so it's not suitable for DOS calculations.

import abipy.abilab as abilab
import abipy.data as abidata

fbnc_kpath = abilab.abiopen(abidata.ref_file("mgb2_kpath_FATBANDS.nc"))

#%%
# Print file info (dimensions, variables ...)
# Note that prtdos = 3, so LM decomposition is not available.

print(fbnc_kpath)

#%%
# Plot the k-points belonging to the path.
fbnc_kpath.ebands.kpoints.plotly()

#%%
# NC files have contributions up to L=4 (g channel)
# but here we are intererested in s,p,d terms only so
# we use the optional argument lmax
lmax = 2

#%%
# Plot the electronic fatbands grouped by atomic type.

fbnc_kpath.plot_fatbands_typeview(lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_typeview(lmax=lmax)

#%%
# Plot the electronic fatbands grouped by L.

fbnc_kpath.plot_fatbands_lview(lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_lview(lmax=lmax)

#%%
# Now we read another FATBANDS file produced on a 18x18x18 k-mesh
# This file can be used to plot the projected-DOS

fbnc_kmesh = abilab.abiopen(abidata.ref_file("mgb2_kmesh181818_FATBANDS.nc"))

#%%
# Let's print the object

print(fbnc_kmesh)
#fbnc_kmesh.ebands.kpoints.plot()

#%%
# Plot the L-PJDOS grouped by atomic type.

fbnc_kmesh.plot_pjdos_typeview(lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kmesh.plotly_pjdos_typeview(lmax=lmax)

#%%
# Plot the L-PJDOS grouped by L.

fbnc_kmesh.plot_pjdos_lview(lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kmesh.plotly_pjdos_lview(lmax=lmax)

#%%
# Now we use the two netcdf files to produce plots with fatbands + PJDOSEs.
# The data for the DOS is taken from pjdosfile.
# sphinx_gallery_thumbnail_number = 6

fbnc_kpath.plot_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                    lmax=lmax, view="type", tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                      lmax=lmax, view="type")

#%%
# fatbands + PJDOS grouped by L

fbnc_kpath.plot_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                    lmax=lmax, view="lview", tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                      lmax=lmax, view="lview")

#fbnc_kpath.close()
#fbnc_kmesh.close()



#!/usr/bin/env python
r"""
Electron DOS
============

This example shows how to compute the DOS from
the eigenvalues stored in the WFK file with the gaussian method.
"""
import abipy.data as abidata
from abipy.abilab import abiopen

# Open the wavefunction file computed with a homogeneous sampling of the BZ
# and extract the band structure on the k-mesh.
with abiopen(abidata.ref_file("si_scf_WFK.nc")) as gs_wfk:
    gs_ebands = gs_wfk.ebands

# Compute the DOS with the Gaussian method (default)
edos = gs_ebands.get_edos(method="gaussian", step=0.01, width=0.1)

#%%
# To plot electron DOS with matplotlib use:
edos.plot(title="Silicon DOS")

#%%
# For the plotly version use:
edos.plotly(title="Silicon DOS")

#%%
# To plot electron DOS and IDOS with matplotlib use:
edos.plot_dos_idos(title="DOS and Integrated DOS")

#%%
# For the plotly version use:
edos.plotly_dos_idos(title="DOS and Integrated DOS")



#!/usr/bin/env python
import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk


def make_scf_input(scf_ngkpt, paral_kgb=0):
    """
    This function constructs the input file for the GS calculation:
    """

    # Initialize MgO structure from abinit variables.
    structure = abilab.Structure.from_abivars(
        acell=3 * [4.252718 * abilab.units.ang_to_bohr],
        rprim=[0.0000000000, 0.5000000000, 0.5000000000,
               0.5000000000, 0.0000000000, 0.5000000000,
               0.5000000000, 0.5000000000, 0.0000000000],
        natom=2,
        ntypat=2,
        typat=[1, 2],
        znucl=[12, 8],
        xred=[0.0000000000, 0.0000000000, 0.0000000000,
              0.5000000000, 0.5000000000, 0.5000000000]
    )

    # Input for GS part.
    # NC pseudos assumed in currect working directory.
    #pseudos = ["Mg-sp-gw.psp8", "O.psp8"]
    pseudos = abidata.pseudos("12mg.pspnc", "O.psp8")

    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    gs_inp.set_vars(
        nband=12,
        paral_kgb=paral_kgb,
        ecut=35.0,        # Too low. Should be ~50
        ngkpt=scf_ngkpt,  # Too coarse
        nshiftk=1,        # Gamma-centered mesh. Important to have the CBM/VBM!
        shiftk=[0, 0, 0],
        tolvrs=1.0e-8,
        diemac=9.0,
        nstep=150,
        nbdbuf=4,
        prtpot=1,        # Print potential for Sternheimer
        #iomode=3,       # Produce output files in netcdf format.
    )

    return gs_inp


def build_flow(options):
    """
    Create a `Flow` for ZPR calculations.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=options.workdir)

    # Build input for the GS calculation and register the first GS ScfTask.
    gs_inp = make_scf_input(scf_ngkpt=[4, 4, 4])
    work = flow.register_scf_task(gs_inp)
    scf_task = work[0]

    # Build input for NSCF calculation along k-path (automatically selected by AbiPy)
    nscf_kpath_inp = gs_inp.make_ebands_input()
    work.register_nscf_task(nscf_kpath_inp, deps={scf_task: "DEN"})

    # This is the ab-initio q-mesh that should be COMMENSURATE with scf_ngkpt
    ddb_ngqpt = [2, 2, 2]

    # Now we build the NSCF tasks with different k-meshes and empty states.
    # Each NSCF task generates one of the WFK files used to build the EPH self-energy.
    # These WFK files are then used to perform convergence tests with respect to
    # the interpolated fine q-mesh.
    # NOTE that in the EPH we are gonna use k_mesh = q_mesh
    ngkpt_fine_list = [
        [8, 8, 8],
        #[12, 12, 12],
        #[32, 32, 32],
    ]

    nscf_empty_states_tasks = []
    for ngkpt_fine in ngkpt_fine_list:
        nscf_empty_kmesh_inp = gs_inp.new_with_vars(
            ngkpt=ngkpt_fine,
            #nband=630,      # Too low. ~300
            #nbdbuf=30,      # Reduces considerably the time needed to converge empty states!
            nband=40,      # Too low. ~300
            nbdbuf=5,      # Reduces considerably the time needed to converge empty states!
            tolwfr=1e-18,
            iscf=-2,
        )
        t = work.register_nscf_task(nscf_empty_kmesh_inp, deps={scf_task: "DEN"})
        nscf_empty_states_tasks.append(t)

    # Create work for phonon calculation on the coarse ddb_ngqpt q-mesh.
    # Electric field and Born effective charges are computed.
    ph_work = flowtk.PhononWork.from_scf_task(scf_task, qpoints=ddb_ngqpt,
                                              is_ngqpt=True, with_becs=True, with_quad=False)
    flow.register_work(ph_work)

    # Build template for e-ph self-energy calculation (real + imag part)
    # The k-points must be in the WFK file
    eph_template = gs_inp.new_with_vars(
        optdriver=7,             # EPH driver.
        eph_task=4,              # Activate computation of the EPH self-energy.
        ddb_ngqpt=ddb_ngqpt,     # Ab-initio q-mesh used to produce the DDB file.
        #nkptgw=2,
        #kptgw=[0, 0, 0,
        #       0.5, 5, 0],
        #bdgw=[1, 8, 1, 8],
        #gw_qprange=1,
        tmesh=[0, 200, 1],    # (start, step, num)
        zcut="0.01 eV",
        # This to speed up the FFT.
        mixprec=1,
        boxcutmin=1.1,
    )

    # Set q-path for Fourier interpolation of phonons.
    eph_template.set_qpath(10)

    # Set q-mesh for phonons DOS.
    eph_template.set_phdos_qmesh(nqsmall=16, method="tetra")

    # Now we use the EPH template to perform a convergence study in which
    # we change the dense q-mesh used to integrate the self-energy and the number of bands.
    # The code will activate the Fourier interpolation of the DFPT potentials
    # if eph_ngqpt_fine != ddb_ngqpt

    # Create empty work to contain EPH tasks with this value of eph_ngqpt_fine
    eph_work = flow.new_work()

    for ngkpt_fine, nscf_task in zip(ngkpt_fine_list, nscf_empty_states_tasks):
        new_inp = eph_template.new_with_vars(
            ngkpt=ngkpt_fine,
            eph_ngqpt_fine=ngkpt_fine,
        ) #, nband=nband)

        # The EPH code requires the GS WFK, the DDB file with all perturbations
        # and the DVDB file with the DFPT potentials (already merged by ph_work)
        deps = {nscf_task: "WFK", ph_work: ["DDB", "DVDB"]}
        eph_work.register_eph_task(new_inp, deps=deps)

    #flow.allocate()
    return flow


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""Optical properties with excitonic effects (Bethe-Salpeter formalism)."""
from __future__ import division, print_function

import os
import numpy as np
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata


def make_scf_nscf_bse_inputs(ngkpt=(6, 6, 6), ecut=6, ecuteps=3,
                             mdf_epsinf=12.0, mbpt_sciss="0.8 eV"):
    """
    Build and returns three `AbinitInput` objects to perform a
    GS-SCF + GS-NSCF + BSE calculation with the model dielectric function.

    Args:
        ngkpt: Three integers giving the number of divisions for the k-mesh.
        ecut: Cutoff energy for the wavefunctions.
        ecuteps: Cutoff energy for the screened interation W_{GG'}.
        mdf_epsinf: Static limit of the macroscopic dielectric functions.
            Used to build the model dielectric function.
        mbpt_sciss: Scissors operator energy (used to open the initial KS gap).
    """
    multi = abilab.MultiDataset(structure=abidata.structure_from_ucell("Si"),
                                pseudos=abidata.pseudos("14si.pspnc"), ndtset=3)
    multi.set_mnemonics(True)

    # Variables common to the three datasets.
    multi.set_vars(
        ecut=ecut,
        nband=8,
        istwfk="*1",
        diemac=12.0,
    )

    # SCF run to get the density.
    multi[0].set_vars(tolvrs=1e-8)
    multi[0].set_kmesh(ngkpt=ngkpt, shiftk=(0, 0, 0))

    # NSCF run on a randomly shifted k-mesh (improve the converge of optical properties)
    multi[1].set_vars(
        iscf=-2,
        nband=15,
        tolwfr=1e-8,
        chksymbreak=0,  # Skip the check on the k-mesh.
    )

    # This shift breaks the symmetry of the k-mesh.
    multi[1].set_kmesh(ngkpt=ngkpt, shiftk=(0.11, 0.21, 0.31))

    # BSE run with Haydock iterative method (only resonant + W + v)
    multi[2].set_vars(
        optdriver=99,                 # BS calculation
        chksymbreak=0,                # To skip the check on the k-mesh.
        bs_calctype=1,                # L0 is contstructed with KS orbitals and energies.
        mbpt_sciss=mbpt_sciss,        # Scissors operator used to correct the KS band structure.
        bs_exchange_term=1,           # Exchange term included.
        bs_coulomb_term=21,           # Coulomb term with model dielectric function.
        mdf_epsinf=mdf_epsinf,        # Parameter for the model dielectric function.
        bs_coupling=0,                # Tamm-Dancoff approximation.
        bs_loband=2,                  # Lowest band included in the calculation
        nband=6,                      # Highest band included in the calculation
        bs_freq_mesh="0 6 0.02 eV",   # Frequency mesh for the dielectric function
        bs_algorithm=2,               # Use Haydock method.
        zcut="0.15 eV",               # Complex shift to avoid divergences in the continued fraction.
        ecutwfn=ecut,                 # Cutoff for the wavefunction.
        ecuteps=ecuteps,              # Cutoff for W and /bare v.
        inclvkb=2,                    # The commutator for the optical limit is correctly evaluated.
    )

    # Same shift as the one used in the previous dataset.
    multi[2].set_kmesh(ngkpt=ngkpt, shiftk=(0.11, 0.21, 0.31))

    scf_input, nscf_input, bse_input = multi.split_datasets()
    return scf_input, nscf_input, bse_input


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    workdir = options.workdir
    if not options.workdir:
        workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=workdir)

    for nk in [4, 6, 8]:
        scf_inp, nscf_inp, bse_inp = make_scf_nscf_bse_inputs(ngkpt=3 * [nk], ecut=6, ecuteps=3)
        work = flowtk.BseMdfWork(scf_inp, nscf_inp, bse_inp)
        flow.register_work(work)

    return flow

@abilab.flow_main
def main(options):
    flow = build_flow(options)
    flow.build_and_pickle_dump()
    return flow


if __name__ == "__main__":
    sys.exit(main())


# coding: utf-8
"""
This module contains objects for postprocessing e-ph calculations
using the results stored in the out_EPH_CUMULANT.nc file.
"""
from __future__ import annotations

import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from scipy.fft import fft, fftfreq,fftshift,ifft
from collections import OrderedDict, namedtuple
from collections.abc import Iterable
from tabulate import tabulate
from monty.string import marquee, list_strings
from monty.functools import lazy_property
from monty.termcolor import cprint
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.core.kpoints import Kpoint, KpointList, Kpath, IrredZone, has_timrev_from_kptopt, find_points_along_path
from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_axlims, set_visible,
    rotate_ticklabels, ax_append_title, set_ax_xylabels, linestyles)
from abipy.tools import duck
from abipy.tools.numtools import gaussian
from abipy.electrons.ebands import ElectronBands, ElectronDos, RobotWithEbands, ElectronBandsPlotter, ElectronDosPlotter
from abipy.abio.robots import Robot
from abipy.eph.common import BaseEphReader
from abipy.eph.sigeph import SigEPhFile,EphSelfEnergy, SigmaPhReader, QpTempState


__all__ = [
    "CumulantQpTempState",
    "CumulantEPhFile",
    "CumulantPhReader",
    "CumulantSelfEnergy",
]


class CumulantQpTempState(QpTempState):
    """
    Quasi-particle result for given (spin, kpoint, band).

    .. Attributes:

        spin: Spin index (C convention, i.e >= 0)
        kpoint: |Kpoint| object.
        band: band index. Global index in the band structure. (C convention, i.e >= 0).
        tmesh: Temperature mesh in Kelvin.
        e0: Initial KS energy.
        qpe: Quasiparticle energy (complex) computed with the linearized perturbative approach (Z factor).
        ze0: Renormalization factor Z computed at e = e0.
        fan0: Fan term (complex) evaluated at e_KS
        dw: Debye-Waller (static, real)
        qpe_oms: Quasiparticle energy (real) in the on-the-mass-shell approximation:
            qpe_oms = e0 + Sigma(e0)

    .. note::

        Energies are in eV.
    """


    def set_energies(self, wmesh, vals_wr, vals_e0ks, ntemp, nwr):
        # Determination of the QP energies
        for it in range(ntemp):
            # Checking where Re Sigma(w) crosses with w => e^QP = e^KS ( set to 0.0 ) + Sigma( e^QP )
            idx = np.argwhere(np.diff(np.sign(vals_wr[it,nwr//4:3*nwr//4].real - wmesh[nwr//4:3*nwr//4] + self.e0))).flatten()
            self.qpe[it] = self.e0 + vals_wr[it,idx]
            # On-the-mass-shell QP energy = e^QP = e^KS ( set to 0.0 ) + Sigma(e^KS)
            self.qpe_oms[it] = self.e0 + vals_e0ks[it]



class CumulantEPhFile(SigEPhFile):
    """
    This file contains the Green's Function using the cumulant expansion from a self-energy eph calculation, the |ElectronBands| on the k-mesh.
    Provides methods to analyze and plot results.

    Usage example:

    .. code-block:: python

        SigmaEphFile = SigEPhFile("out_SIGEPH.nc")
        with CumulantEPhFile("out_EPH_CUMULANT.nc", SigmaEphFile) as ncfile:
            print(ncfile)
            ncfile.ebands.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: CumulantEPhFile
    """

    @classmethod
    def from_file(cls, filepath: str) -> CumulantEPhFile:
        """Initialize the object from a netcdf file."""
        return cls(filepath)


    def __init__(self, filepath):
        self._filepath = filepath
        self.reader = r = CumulantPhReader(filepath)
        self.nkcalc = r.nkcalc
        self.ntemp = r.ntemp
        self.nqbz = r.nqbz
        self.nqibz = r.nqibz
        self.ngqpt = r.ngqpt


        self.symsigma = -1


    def __str__(self):
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        ##app("")
        ##app(self.structure.to_string(verbose=verbose, title="Structure"))
        #app("")
        #app(self.ebands.to_string(with_structure=False, verbose=verbose, title="KS Electron Bands"))
        app("")
        # SigmaEPh section.
        app(marquee("CumulantEPh calculation", mark="="))
        app("Number of k-points in Sigma_{nk}: %d" % (self.nkcalc))
        # These variables have added recently
        sigma_ngkpt = self.reader.read_value("sigma_ngkpt", default=None)
        sigma_erange = self.reader.read_value("sigma_erange", default=None)
        #dvdb_add_lr = self.reader.read_value("dvdb_add_lr", default=None)
        #app("sigma_ngkpt: %s, sigma_erange: %s" % (sigma_ngkpt, sigma_erange))
        app("Max bstart: %d, min bstop: %d" % (self.reader.max_bstart, self.reader.min_bstop))
        app("Initial ab-initio q-mesh:\n\tngqpt: %s, with nqibz: %s" % (str(self.ngqpt), self.nqibz))
        #eph_ngqpt_fine = self.reader.read_value("eph_ngqpt_fine")
        #if np.all(eph_ngqpt_fine == 0): eph_ngqpt_fine = self.ngqpt
        #app("q-mesh for self-energy integration (eph_ngqpt_fine): %s" % (str(eph_ngqpt_fine)))
        #app("k-mesh for electrons:")
        #app("\t" + self.ebands.kpoints.ksampling.to_string(verbose=verbose))
        #app("Number of bands included in e-ph self-energy sum: %d" % (self.nbsum))
        #app("zcut: %.5f (Ha), %.3f (eV)" % (self.zcut, self.zcut * abu.Ha_eV))
        app("Number of temperatures: %d, from %.1f to %.1f (K)" % (self.ntemp, self.tmesh[0], self.tmesh[-1]))
        #app("symsigma: %s" % (self.symsigma))
        #app("Has Eliashberg function: %s" % (self.has_eliashberg_function))
        #app("Has Spectral function: %s" % (self.has_spectral_function))

        # Build Table with direct gaps. Only the results for the first and the last T are shown if not verbose.
        #if verbose:
        #    it_list = list(range(self.ntemp))
        #else:
        #    it_list = [0, -1] if self.ntemp != 1 else [0]
        #app("\nPrinting QP results for %d temperatures. Use --verbose to print all results." % len(it_list))

        # QP corrections
        #for it in it_list:
        #    app("\nKS, QP (Z factor) and on-the-mass-shell (OTMS) direct gaps in eV for T = %.1f K:" % self.tmesh[it])
        #    data = []
        #    for ikc, kpoint in enumerate(self.sigma_kpoints):
        #        for spin in range(self.nsppol):
        #            ks_gap = self.ks_dirgaps[spin, ikc]
        #            qp_gap = self.qp_dirgaps_t[spin, ikc, it]
        #            oms_gap = self.qp_dirgaps_otms_t[spin, ikc, it]
        #            data.append([spin, repr(kpoint), ks_gap, qp_gap, qp_gap - ks_gap, oms_gap, oms_gap - ks_gap])
        #    app(str(tabulate(data,
        #        headers=["Spin", "k-point", "KS_gap", "QPZ0_gap", "QPZ0 - KS", "OTMS_gap", "OTMS - KS"],
        #        floatfmt=".3f")))
        #    app("")
        #else:
        # Print info on Lifetimes?

        #if verbose > 1:
        #    app("K-points and bands included in self-energy corrections:")
        #    for spin in range(self.nsppol):
        #        for ikc, kpoint in enumerate(self.sigma_kpoints):
        #            post = "ikc: %d" % (ikc if self.nsppol == 1 else "ikc: %d, spin: %d" % (ikc, spin))
        #            app("\t%s: bstart: %d, bstop: %d, %s" % (
        #                repr(kpoint), self.bstart_sk[spin, ikc], self.bstop_sk[spin, ikc], post))

        return "\n".join(lines)

    def get_cumulant_skb(self, spin, kpoint, band):
        """"Return e-ph self-energy for the given (spin, kpoint, band)."""
        return self.reader.read_cumulant_skb(spin, kpoint, band)




class CumulantPhReader(SigmaPhReader):
    """
    Reads data from file and constructs objects.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: CumulantPhReader
    """

    def __init__(self, path: str):
        super().__init__(path)

        # Check if the cumulant function exists
        # It is an optional output, mostly for debuging
        # If it exists, then G(t) and time_mesh also exists
        variables = self.read_varnames()
        if "ct_vals" in variables:
            self.ct_vals_exists = True
        else:
            self.ct_vals_exists = False

    def read_cumulant_skb(self, spin, kpoint, band):
        """
        Returns the e-ph self energy for the given (spin, k-point, band).

        Args:
            spin: Spin index
            kpoint: K-point in self-energy. Accepts |Kpoint|, vector or index.
            band: band index.

        Return: :class:`EphSelfEnergy` object.
        """
        if self.nwr == 0:
            raise ValueError("%s does not contain spectral function data." % self.path)

        spin, ikc, ib, kpoint = self.get_sigma_skb_kpoint(spin, kpoint, band)

        # Abinit fortran (Ha units)
        # real(dp) :: wrmesh_b(nwr, max_nbcalc, nkcalc, nsppol)
        # Frequency mesh along the real axis (Ha units) used for the different bands
        wmesh = self.read_variable("wrmesh_b")[spin, ikc, ib, :] * abu.Ha_eV

        # Set time mesh, cumulant function and green's function in time domain as None ( default )
        time_mesh =ct_vals = gt_vals = None
        if self.ct_vals_exists:
            # For debugging proposes read time mesh, cumulant function and green's function in time domain

            # real(dp) :: time_mesh(nwr, ntemp, max_nbcalc, nkcalc, nsppol)
            # Time mesh (ps units)
            time_mesh = self.read_variable("time_mesh")[spin,ikc,ib,:,:]
            
            # complex(dpc) :: ct_vals(complex, nwr, ntemp, max_nbcalc, nkcalc, nsppol)
            # Cumulant function with respect to time
            ct_vals = self.read_variable("ct_vals")[spin,ikc,ib,:,:,:]
            ct_vals = ct_vals[:, :, 0] + 1j* ct_vals[:, :, 1]

            # complex(dpc) :: gt_vals(complex, nwr, ntemp, max_nbcalc, nkcalc, nsppol)
            # Green's function with respect to time
            gt_vals = self.read_variable("gt_vals")[spin,ikc,ib,:,:,:]
            gt_vals = gt_vals[:,:, 0] + 1j*gt_vals[:, :, 1]

        # complex(dpc) :: gw_vals(nwr, ntemp, max_nbcalc, nkcalc, nsppol)
        # Green's function in frequency domain
        gw_vals = self.read_variable("gw_vals")[spin,ikc,ib,:,:,:] / abu.Ha_eV
        gw_vals = gw_vals[:, :, 0] + 1j * gw_vals[:, :, 1]

        # Spectral function obtained from the Gren's function in frequency domain
        spfunccumul_wr = -1.0*gw_vals[:, :].imag/np.pi

        # Read QP data. Note band instead of ib index.
        qp = self.read_qp(spin, ikc, band)

        return CumulantSelfEnergy(wmesh, qp, gw_vals, spfunccumul_wr, time_mesh = time_mesh, ct_vals= ct_vals, gt_vals= gt_vals)


    def read_qp(self, spin, kpoint, band, ignore_imag=False):
        """
        Return :class:`QpTempState` for the given (spin, kpoint, band)
        (NB: band is a global index i.e. unshifted)
        Only real part is returned if ``ignore_imag``.
        """
        spin, ikc, ibc, kpoint = self.get_sigma_skb_kpoint(spin, kpoint, band)

        # Quasi-particle energies set initially to zero
        qpe = np.zeros((self.ntemp),dtype = complex)
        qpe_oms = np.zeros((self.ntemp))

        # Fan and Debbye-Waller self-energies set to zero
        fan0 = np.zeros((self.ntemp),dtype = complex)
        dw = np.zeros((self.ntemp))

        # Renormalization of the Quasi-Particle set to zero
        ze0 = 0.0

        # real(dp) :: ks_enes( max_nbcalc, nkcalc, nsppol )
        # Kohn-Sham eigenvalues
        e0 = self.read_variable("ks_enes")[spin, ikc, ibc] * abu.Ha_eV


        return CumulantQpTempState(spin=spin, kpoint=kpoint, band=band, tmesh=self.tmesh,
                           e0=e0, qpe = qpe, ze0=ze0, fan0=fan0, dw=dw, qpe_oms = qpe_oms)






class CumulantSelfEnergy(EphSelfEnergy):
    

    def __init__(self, wmesh, qp, gw_vals, spfunccumul_wr, time_mesh = None, ct_vals = None, gt_vals = None, 
                 vals_e0ks=None, dvals_de0ks=None, dw_vals=None,
                 frohl_vals_e0ks=None, frohl_dvals_de0ks=None, frohl_spfunc_wr=None):
        
        # Set dimensions
        ntemp = len(qp.tmesh)
        nwr = len(wmesh)

        # In case of debugging, seting:
        # Green's function in time domain, cumulant function and time mesh
        self.gt_vals = gt_vals
        self.ct_vals = ct_vals
        self.time_mesh = time_mesh

        # Set Debye-Waller to zero, probably is the same as Dyson-Migdal since it is added 
        # separately from the cumulant calculation
        dw_vals = np.zeros((ntemp))

        # Calculation of the self-energy (vals_wr)
        # Calculation of the self-energy evaluated at the KS energy (vals_e0ks)
        # Calculation of the derivative of the self-energy evaluated at the KS energy (dvals_de0ks)
        self.gw = gw_vals
        vals_wr, vals_e0ks, dvals_de0ks = self.calculate_sigma_skb_fromgw(wmesh, qp, gw_vals)

        # Setting spectral function and Gre 
        self.spfunccumul_wr = spfunccumul_wr

        # Calculating QP energies
        qp.set_energies(wmesh, vals_wr, vals_e0ks, ntemp, nwr)

        super().__init__(wmesh, qp, vals_e0ks, dvals_de0ks, dw_vals, vals_wr, spfunccumul_wr,
                   frohl_vals_e0ks=None, frohl_dvals_de0ks=None, frohl_spfunc_wr=None)


    def calculate_sigma_skb_fromgw(self, wmesh, qp, gw_vals):

        # Initial setting
        nwr = len(wmesh)
        ntemp = len(qp.tmesh)
        sigma = np.zeros((ntemp,nwr),dtype=complex)
        e0ks = np.zeros((ntemp))
        de0ks = np.zeros((ntemp))
        
        # Identify the index of the KS energy is located ( probably at nwr//2 )
        idx_e0 = np.argmin( np.abs(wmesh - qp.e0))

        for it in range(ntemp):
            # Use Dyson equations to determine self-energy, only good for energies close to KS
            sigma[it,:] = wmesh[:]- qp.e0 - 1/gw_vals[it,:]
            # Evaluate the self-energy at KS energy
            e0ks[it] = sigma[it,idx_e0].real
            # The derivative of the self-energy evaluated at the KS energy.
            de0ks[it] = np.gradient(sigma[it,:].real, wmesh[1]-wmesh[0])[idx_e0]
        return sigma, e0ks, de0ks

    @classmethod
    def calculate_gw_from_sigeph_skb(cls, sigeph, time_tol = 1e-4):
        
        # Initialization
        wmesh_init = sigeph.wmesh
        sigma = sigeph.vals_wr
        sigma0 = sigeph.vals_e0ks.real
        e0 = sigeph.qp.e0
        nwr = sigeph.nwr
        ntemp = sigeph.ntemp
        qpe = np.zeros((ntemp),dtype = complex)
        qpe_oms = np.zeros((ntemp))
        fan0 = np.zeros((ntemp),dtype = complex)
        dw = np.zeros((ntemp))
        ze0 = 0.0
        qp = CumulantQpTempState(spin=sigeph.spin, kpoint=sigeph.kpoint, band=sigeph.band, tmesh=sigeph.tmesh,
                           e0=sigeph.qp.e0, qpe = qpe, ze0=ze0, fan0=fan0, dw=dw, qpe_oms = qpe_oms)

       
        # Frequency mesh step
        w_step = wmesh_init[1] - wmesh_init[0]
        # Shift frequency mesh to set the KS energy close to 0.0, with principal value offseting the mesh slightly.
        wmesh = wmesh_init - e0 - 0.5 * w_step
        # Create the time mesh
        t = np.linspace(0.0,1,nwr)*2*np.pi/w_step
        # Time mesh step
        t_step = t[1] - t[0]
        gw = np.zeros((ntemp,nwr), dtype=complex)
    
        for itemp in range(ntemp):
            #time_max = np.log(time_tol)/ ( -1.0 * np.abs(sigeph.vals_e0ks[itemp].imag))
            #Give warning if time_max < 2*np.pi/w_step
            
         
            beta = np.abs( sigma[itemp].imag)/np.pi

            # Calculate the part of the cumulant function that can be Fourier Transformed
            c1= fft((beta/(wmesh**2)),nwr)*w_step

            # The odd indexes give negative values
            c1[1::2] = -1.0 * c1[1::2]

            # Other terms of the cumulant function
            c2 = - 1j * t * sigma0
            c3 = -1.0 * np.trapz(beta/wmesh**2,x=wmesh) * np.ones(nwr)

            # Calculation of the cumulant function
            Ct = c1 + c2 + c3

            # Green's Function in time domain calculated with the cumulant function
            Gt = -1j * np.exp(Ct)

            # Green's function in frequency domain
            gw_temp = fftshift(ifft(Gt,nwr))*nwr*t_step
            R = 0.5* Gt[0]*t_step
            gw[itemp,:] = gw_temp - R

        # Spectral Function
        spfunc = -1.0 * gw.imag / np.pi


        return cls( wmesh_init, qp, gw, spfunc, time_mesh = t, ct_vals = c1+c2+c3, gt_vals = Gt )





# coding: utf-8
"""
Flows for electron-phonon calculations (high-level interface)
"""
from __future__ import annotations

import numpy as np

from abipy.core.kpoints import kpath_from_bounds_and_ndivsm
from abipy.abio.inputs import AbinitInput
from .nodes import Node
from .works import Work, PhononWork, PhononWfkqWork
from .flows import Flow


class EphPotFlow(Flow):
    r"""
    This flow computes the e-ph scattering potentials on a q-mesh defined by ngqpt
    and a list of q-points (usually a q-path) specified by the user.
    The DFPT potentials on the q-mesh are merged in the DVDB located in the outdata
    of the second work while the DFPT potentials on the q-path are merged in the DVDB
    located in the outdata of the third work.
    These DVDB files are then passed to the EPH code to compute the average over the unit
    cell of the periodic part of the scattering potentials as a function of q.
    Results are stored in the V1QAVG.nc files of the outdata of the tasks in the fourth work.
    """

    @classmethod
    def from_scf_input(cls, workdir: str, scf_input: AbinitInput,
                       ngqpt, qbounds,
                       ndivsm=5, with_becs=True, with_quad=True, dvdb_add_lr_list=(0, 1, 2),
                       ddb_filepath=None, dvdb_filepath=None,
                       ddk_tolerance=None, prepgkk=0, manager=None) -> EphPotFlow:
        """
        Build the flow from an input file representing a GS calculation.

        Args:
            workdir: Working directory.
            scf_input: Input for the GS SCF run.
            ngqpt: 3 integers defining the q-mesh.
            qbounds: List of boundaries defining the q-path used for the computation of the GKQ files.
                The q-path is automatically generated using `ndivsm` and the reciprocal-space metric.
                If `ndivsm` is 0, the code assumes that `qbounds` contains the full list of q-points
                and no pre-processing is performed.
            ndivsm: Number of points in the smallest segment of the path defined by `qbounds`.
                Use 0 to pass full list of q-points.
            with_becs: Activate calculation of Electric field and Born effective charges.
            with_quad: Activate calculation of dynamical quadrupoles. Require `with_becs`
                Note that only selected features are compatible with dynamical quadrupoles.
                Please consult <https://docs.abinit.org/topics/longwave/>
            dvdb_add_lr_list: List of dvdb_add_lr values to consider in the interpolation.
            ddb_filepath, dvdb_filepath: Paths to the DDB/DVDB files that will be used
                to bypass the DFPT computation on the `ngqpt` mesh.
            ddk_tolerance: dict {"varname": value} with the tolerance used in the DDK run if `with_becs`.
            prepgkk: 1 to activate computation of all 3 * natom perts (debugging option).
            manager: |TaskManager| object.
        """
        flow = cls(workdir=workdir, manager=manager)

        # First work with GS run.
        scf_task = flow.register_scf_task(scf_input)[0]

        if dvdb_filepath or ddb_filepath:
            # Use input files to bypass computation of work_qmesh.
            if not (dvdb_filepath and ddb_filepath):
                raise ValueError("Both dvdb_filepath and ddb_filepath must be specified.")
            work_qmesh = None
            # TODO Should check that ddb.qmesh == ngqpt
            ddb_node = Node.as_node(ddb_filepath)
            dvdb_node = Node.as_node(dvdb_filepath)

        else:
            # Second work to compute phonons on the input nqgpt q-mesh.
            work_qmesh = PhononWork.from_scf_task(scf_task, qpoints=ngqpt, is_ngqpt=True,
                                                  with_becs=with_becs, with_quad=with_quad,
                                                  ddk_tolerance=ddk_tolerance)
            flow.register_work(work_qmesh)

        if ndivsm > 0:
            # Generate list of q-points from qbounds and ndivsm.
            qpath_list = kpath_from_bounds_and_ndivsm(qbounds, ndivsm, scf_input.structure)
        elif ndivsm == 0:
            # Use input list of q-points.
            qpath_list = np.reshape(qbounds, (-1, 3))
        else:
            raise ValueError("ndivsm cannot be negative. Received ndivsm: %s" % ndivsm)

        # Third Work: compute WFK/WFQ and phonons for qpt in qpath_list.
        # Don't include BECS because they have been already computed in the previous work.
        work_qpath = PhononWfkqWork.from_scf_task(
                       scf_task, qpath_list, ph_tolerance=None, tolwfr=1.0e-22, nband=None,
                       with_becs=False, ddk_tolerance=None, shiftq=(0, 0, 0), is_ngqpt=False, remove_wfkq=True,
                       prepgkk=prepgkk, manager=manager)

        flow.register_work(work_qpath)

        # Now we compute matrix elements fully ab-initio for each q-point.
        eph_work = Work()

        for eph_task in (-15, 15):
            eph_inp = scf_input.new_with_vars(
                optdriver=7,
                ddb_ngqpt=ngqpt,    # q-mesh associated to the DDB file.
                #dvdb_ngqpt=ngqpt,  # q-mesh associated to the DDVDB file.
                prtphdos=0,
                eph_task=eph_task
            )

            if eph_task == -15:
                # Use DVDB with ab-initio POTS along q-path to produce V1QAVG
                if work_qmesh is not None:
                    deps = {work_qmesh: "DDB", work_qpath: "DVDB"}
                else:
                    deps = {ddb_node: "DDB", work_qpath: "DVDB"}

                eph_work.register_eph_task(eph_inp, deps=deps)

            elif eph_task == 15:
                # Use q-mesh to interpolate along the same q-path as above.
                # use dvdb_add_lr to deactivate/activate the treatment of the LR part.
                if work_qmesh is not None:
                    deps = {work_qmesh: ["DDB", "DVDB"]}
                else:
                    deps = {ddb_node: "DDB", dvdb_node: "DVDB"}

                for dvdb_add_lr in dvdb_add_lr_list:
                    new_inp = eph_inp.new_with_vars(dvdb_add_lr=dvdb_add_lr, ph_qpath=qpath_list)
                    eph_work.register_eph_task(new_inp, deps=deps)

        flow.register_work(eph_work)

        return flow


class GkqPathFlow(Flow):
    r"""
    This flow computes the gkq e-ph matrix elements <k+q|\Delta V_q|k> for a list of q-points (usually a q-path).
    The results are stored in the GKQ.nc files for the different q-points. These files can be used to analyze the behaviour
    of the e-ph matrix elements as a function of qpts with the objects provided by the abipy.eph.gkq module.
    It is also possible to compute the e-ph matrix elements using the interpolated DFPT potentials
    if test_ft_interpolation is set to True.
    """

    @classmethod
    def from_scf_input(cls, workdir: str, scf_input: AbinitInput,
                       ngqpt, qbounds,
                       ndivsm=5, with_becs=True, with_quad=True, dvdb_add_lr_list=(0, 1, 2),
                       ddb_filepath=None, dvdb_filepath=None,
                       ddk_tolerance=None, test_ft_interpolation=False, prepgkk=0, manager=None) -> GkqPathFlow:
        """
        Build the flow from an input file representing a GS calculation.

        Args:
            workdir: Working directory.
            scf_input: Input for the GS SCF run.
            ngqpt: 3 integers defining the q-mesh.
            qbounds: List of boundaries defining the q-path used for the computation of the GKQ files.
                The q-path is automatically generated using `ndivsm` and the reciprocal-space metric.
                If `ndivsm` is 0, the code assumes that `qbounds` contains the full list of q-points
                and no pre-processing is performed.
            ndivsm: Number of points in the smallest segment of the path defined by `qbounds`.
                Use 0 to pass list of q-points.
            with_becs: Activate calculation of Electric field and Born effective charges.
            with_quad: Activate calculation of dynamical quadrupoles. Require `with_becs`
                Note that only selected features are compatible with dynamical quadrupoles.
                Please consult <https://docs.abinit.org/topics/longwave/>
            dvdb_add_lr_list: List of dvdb_add_lr values to consider in the interpolation.
            ddb_filepath, dvdb_filepath: Paths to the DDB/DVDB files that will be used
                to bypass the DFPT computation on the `ngqpt` mesh.
            ddk_tolerance: dict {"varname": value} with the tolerance used in the DDK run if `with_becs`.
            test_ft_interpolation: True to add an extra Work in which the GKQ files are computed
                using the interpolated DFPT potentials and the q-mesh defined by `ngqpt`.
                The quality of the interpolation depends on the convergence of the BECS, epsinf and `ngqpt`.
                and the treatment of the LR part of the e-ph scattering potentials.
            prepgkk: 1 to activate computation of all 3 * natom perts (debugging option).
            manager: |TaskManager| object.
        """
        flow = cls(workdir=workdir, manager=manager)

        # First work with GS run.
        scf_task = flow.register_scf_task(scf_input)[0]

        if dvdb_filepath or ddb_filepath:
            # Use input files to bypass computation of work_qmesh.
            if not (dvdb_filepath and ddb_filepath):
                raise ValueError("Both dvdb_filepath and ddb_filepath must be specified.")
            work_qmesh = None
            # TODO Should check that ddb.qmesh == ngqpt
            ddb_node = Node.as_node(ddb_filepath)
            dvdb_node = Node.as_node(dvdb_filepath)

        else:

            # Second work to compute phonons on the input nqgpt q-mesh.
            work_qmesh = PhononWork.from_scf_task(scf_task, qpoints=ngqpt, is_ngqpt=True,
                                                  with_becs=with_becs, with_quad=with_quad,
                                                  ddk_tolerance=ddk_tolerance)
            flow.register_work(work_qmesh)

        if ndivsm > 0:
            # Generate list of q-points from qbounds and ndivsm.
            qpath_list = kpath_from_bounds_and_ndivsm(qbounds, ndivsm, scf_input.structure)
        elif ndivsm == 0:
            # Use input list of q-points.
            qpath_list = np.reshape(qbounds, (-1, 3))
        else:
            raise ValueError("ndivsm cannot be negative. Received ndivsm: %s" % ndivsm)

        # Third Work. Compute WFK/WFQ and phonons for qpt in qpath_list.
        # Don't include BECS because they have been already computed in the previous work.
        work_qpath = PhononWfkqWork.from_scf_task(
                       scf_task, qpath_list, ph_tolerance=None, tolwfr=1.0e-22, nband=None,
                       with_becs=False, ddk_tolerance=None, shiftq=(0, 0, 0), is_ngqpt=False, remove_wfkq=False,
                       prepgkk=prepgkk, manager=manager)

        flow.register_work(work_qpath)

        def make_eph_input(scf_inp, ngqpt, qpt):
            """
            Build input file to compute GKQ.nc file from GS SCF input.
            The calculation requires GS wavefunctions WFK, WFQ, a DDB file and a DVDB file
            """
            return scf_inp.new_with_vars(
                optdriver=7,
                eph_task=-2,
                nqpt=1,
                qpt=qpt,
                ddb_ngqpt=ngqpt,  # q-mesh associated to the DDB file.
                prtphdos=0,
            )

        # Now we compute e-ph matrix elements fully ab-initio for each q-point.
        eph_work = Work()

        qseen = set()
        for task in work_qpath.phonon_tasks:
            qpt = tuple(task.input["qpt"])
            if qpt in qseen: continue
            qseen.add(qpt)
            t = eph_work.register_eph_task(make_eph_input(scf_input, ngqpt, qpt), deps=task.deps)

            if work_qmesh is not None:
                t.add_deps({work_qmesh: "DDB", work_qpath: "DVDB"})
            else:
                t.add_deps({ddb_node: "DDB", work_qpath: "DVDB"})

        flow.register_work(eph_work)

        # Here we build another work to compute the gkq matrix elements
        # with interpolated potentials along the q-path.
        # The potentials are interpolated using the input ngqpt q-mesh.
        if test_ft_interpolation:

            for dvdb_add_lr in dvdb_add_lr_list:
                inteph_work = Work()
                qseen = set()
                for task in work_qpath.phonon_tasks:
                    qpt = tuple(task.input["qpt"])
                    if qpt in qseen: continue
                    qseen.add(qpt)
                    eph_inp = make_eph_input(scf_input, ngqpt, qpt)
                    # Note eph_use_ftinterp 1 to force the interpolation
                    # of the DFPT potentials with eph_task -2.
                    eph_inp.set_vars(eph_use_ftinterp=1, dvdb_add_lr=dvdb_add_lr)
                    t = inteph_work.register_eph_task(eph_inp, deps=task.deps)

                    if work_qmesh is not None:
                        t.add_deps({work_qmesh: ["DDB", "DVDB"]})
                    else:
                        t.add_deps({ddb_node: "DDB", dvdb_node: "DVDB"})

                flow.register_work(inteph_work)

        return flow
